#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
BeamlineTestMockup

"""

import logging
import os
import tempfile
from datetime import datetime

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects import SimpleHTML

__credits__ = ["MXCuBE collaboration"]


TEST_DICT = {
    "example_one": "Beamline test example no 1.",
    "example_two": "Beamline test example no 2.",
    "file_system": "File system",
}

TEST_COLORS_TABLE = {False: "#FFCCCC", True: "#CCFFCC"}
TEST_COLORS_FONT = {False: "#FE0000", True: "#007800"}


class BeamlineTestMockup(HardwareObject):
    """BeamlineTestMockup"""

    def __init__(self, name):
        """init"""

        HardwareObject.__init__(self, name)

        self.ready_event = None
        self.test_queue_dict = None
        self.current_test_procedure = None
        self.beamline_name = None
        self.test_directory = None
        self.test_source_directory = None
        self.test_filename = None

        self.available_tests_dict = {}
        self.startup_test_list = []
        self.results_list = None
        self.results_html_list = None

    def init(self):
        """init"""

        self.ready_event = gevent.event.Event()

        self.beamline_name = HWR.beamline.session.beamline_name

        self.test_directory = self.get_property("results_directory")
        if self.test_directory is None:
            self.test_directory = os.path.join(
                tempfile.gettempdir(), "mxcube", "beamline_test"
            )
            logging.getLogger("HWR").debug(
                "BeamlineTest: directory for test "
                "reports not defined. Set to: %s" % self.test_directory
            )
        self.test_source_directory = os.path.join(
            self.test_directory, datetime.now().strftime("%Y_%m_%d_%H") + "_source"
        )

        self.test_filename = "mxcube_test_report"

        try:
            for test in eval(self.get_property("available_tests")):
                self.available_tests_dict[test] = TEST_DICT[test]
        except Exception:
            logging.getLogger("HWR").debug(
                "BeamlineTest: Available tests are "
                + "not defined in xml. Setting all tests as available."
            )
        if self.available_tests_dict is None:
            self.available_tests_dict = TEST_DICT

        try:
            self.startup_test_list = eval(self.get_property("startup_tests"))
        except Exception:
            logging.getLogger("HWR").debug("BeamlineTest: Test list not defined.")

        if self.get_property("run_tests_at_startup") is True:
            self.start_test_queue(self.startup_test_list)

    def start_test_queue(self, test_list, create_report=True):
        """
        Descrip. :
        """
        if create_report:
            try:
                logging.getLogger("HWR").debug(
                    "BeamlineTest: Creating directory %s" % self.test_directory
                )
                if not os.path.exists(self.test_directory):
                    os.makedirs(self.test_directory)

                logging.getLogger("HWR").debug(
                    "BeamlineTest: Creating source directory %s"
                    % self.test_source_directory
                )
                if not os.path.exists(self.test_source_directory):
                    os.makedirs(self.test_source_directory)
            except Exception:
                logging.getLogger("HWR").warning(
                    "BeamlineTest: Unable to create test directories"
                )
                return

        self.results_list = []
        self.results_html_list = []
        for test_index, test_name in enumerate(test_list):
            test_method_name = "test_" + test_name.lower()
            if hasattr(self, test_method_name):
                if test_name in TEST_DICT:
                    logging.getLogger("HWR").debug(
                        "BeamlineTest: Executing test %s (%s)"
                        % (test_name, TEST_DICT[test_name])
                    )

                    progress_info = {
                        "progress_total": len(test_list),
                        "progress_msg": "executing %s" % TEST_DICT[test_name],
                    }
                    self.emit("testProgress", (test_index, progress_info))

                    start_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    self.current_test_procedure = gevent.spawn(
                        getattr(self, test_method_name)
                    )
                    test_result = self.current_test_procedure.get()

                    # self.ready_event.wait()
                    # self.ready_event.clear()
                    end_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    self.results_list.append(
                        {
                            "short_name": test_name,
                            "full_name": TEST_DICT[test_name],
                            "result_bit": test_result.get("result_bit", False),
                            "result_short": test_result.get("result_short", ""),
                            "start_time": start_time,
                            "end_time": end_time,
                        }
                    )

                    self.results_html_list.append(
                        "<h2 id=%s>%s</h2>" % (test_name, TEST_DICT[test_name])
                    )
                    self.results_html_list.append("Started: %s<br>" % start_time)
                    self.results_html_list.append("Ended: %s<br>" % end_time)
                    if test_result.get("result_short"):
                        self.results_html_list.append(
                            "<h3><font color=%s>Result : %s</font></h3>"
                            % (
                                TEST_COLORS_FONT[test_result["result_bit"]],
                                test_result["result_short"],
                            )
                        )
                    if len(test_result.get("result_details", [])) > 0:
                        self.results_html_list.append("<h3>Detailed results:</h3>")
                        self.results_html_list.extend(
                            test_result.get("result_details", [])
                        )
            else:
                msg = (
                    "<h2><font color=%s>Execution method %s "
                    + "for the test %s does not exist</font></h3>"
                )
                self.results_html_list.append(
                    msg
                    % (TEST_COLORS_FONT[False], test_method_name, TEST_DICT[test_name])
                )
                logging.getLogger("HWR").error(
                    "BeamlineTest: Test method %s not available" % test_method_name
                )
            self.results_html_list.append("</p>\n<hr>")

        # html_filename = None
        if create_report:
            self.generate_html_report()

        # self.emit('testFinished', html_filename)

    def test_example_one(self):
        """Text one"""
        result = {}

        current_energy = HWR.beamline.energy.get_value()

        result["result_bit"] = current_energy < 12
        result["result_short"] = "Test passed (energy = %.2f)" % current_energy
        result["result_details"] = ["An example test that was successful"]
        self.ready_event.set()
        return result

    def test_example_two(self):
        """Text one"""

        result = {}
        result["result_bit"] = False
        result["result_short"] = "Test failed"
        result["result_details"] = ["An example test that failed"]
        self.ready_event.set()
        return result

    def get_available_tests(self):
        """
        Descript. :
        """
        return self.available_tests_dict

    def get_startup_test_list(self):
        """
        Descript. :
        """
        test_list = []
        for test in self.startup_test_list:
            if TEST_DICT.get(test):
                test_list.append(TEST_DICT[test])
        return test_list

    def generate_html_report(self):
        """
        Descript. :
        """
        html_filename = os.path.join(self.test_directory, self.test_filename + ".html")
        # pdf_filename = os.path.join(self.test_directory, self.test_filename + ".pdf")
        archive_filename = os.path.join(
            self.test_directory,
            datetime.now().strftime("%Y_%m_%d_%H") + "_" + self.test_filename,
        )

        try:
            output_file = open(html_filename, "w")
            output_file.write(SimpleHTML.create_html_start("Beamline test summary"))
            output_file.write("<h1>Beamline %s Test results</h1>" % self.beamline_name)

            output_file.write("<h2>Executed tests:</h2>")
            table_cells = []
            for test in self.results_list:
                table_cells.append(
                    [
                        "bgcolor=%s" % TEST_COLORS_TABLE[test["result_bit"]],
                        "<a href=#%s>%s</a>" % (test["short_name"], test["full_name"]),
                        test["result_short"],
                        test["start_time"],
                        test["end_time"],
                    ]
                )

            table_rec = SimpleHTML.create_table(
                ["Name", "Result", "Start time", "End time"], table_cells
            )
            for row in table_rec:
                output_file.write(row)
            output_file.write("\n<hr>\n")

            for test_result in self.results_html_list:
                output_file.write(test_result + "\n")

            output_file.write(SimpleHTML.create_html_end())
            output_file.close()

            self.emit("htmlGenerated", html_filename)
            logging.getLogger("HWR").info(
                "BeamlineTest: Test result written in file %s" % html_filename
            )
        except Exception:
            logging.getLogger("HWR").error(
                "BeamlineTest: Unable to generate html report file %s" % html_filename
            )

        # try:
        #    pdfkit.from_url(html_filename, pdf_filename)
        #    logging.getLogger("GUI").info("PDF report %s generated" % pdf_filename)
        # except Exception:
        #    logging.getLogger("GUI").info(
        #        "Unable to generate PDF report %s" % pdf_filename
        #    )

        self.emit("testFinished", html_filename)

    def get_result_html(self):
        """
        Descript. :
        """
        html_filename = os.path.join(self.test_directory, self.test_filename + ".html")
        if os.path.exists(html_filename):
            return html_filename



from mxcubecore.BaseHardwareObjects import HardwareObject


class RobodiffController(HardwareObject):
    def __init__(self, *args):
        HardwareObject.__init__(self, *args)

    def init(self, *args):
        self.__controller = self.get_object_by_role("controller").robot

    def set_diagfile(self, diagfile):
        self.__controller.diagfile = diagfile

    def __getattr__(self, attr):
        if attr.startswith("__"):
            raise AttributeError(attr)
        return getattr(self.__controller, attr)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import sys
import time

from PIL import Image

from mxcubecore.BaseHardwareObjects import HardwareObject

sys.path.append("/opt/dectris/albula/4.0/python/")


try:
    from dectris import albula
except:
    print("albula module not found")


class P11AlbulaView(HardwareObject):
    default_interval = 0.5  # secs
    stoptimer = -1.0

    def init(self):
        self.alive = False
        self.stream = False
        self.viewer = None
        self.subframe = None

        interval = self.get_property("update_interval")

        self.interval = interval is not None and interval or self.default_interval

        self.mask_4m = None
        self.mask_16m = None

        mask_path_4m = self.get_property("mask_4m")
        mask_path_16m = self.get_property("mask_16m")

        self.eigerMonitor = None
        self.eigerThread = None

        if parent is not None and hasattr(parent, "eigerThread"):
            try:
                prop = parent.eigerThread.proxyMonitor.get_property("Host")["Host"]
            except:
                pass
            else:
                self.eigerMonitor = albula.DEigerMonitor(prop[0])
                self.eigerThread = parent.eigerThread

        try:
            mask_tif_4m = Image.open(mask_path_4m)
            mask_tif_16m = Image.open(mask_path_16m)
        except:
            pass
        else:
            mask_array_4m = np.asarray(mask_tif_4m)
            mask_array_16m = np.asarray(mask_tif_16m)

            self.mask_4m = albula.DImage(mask_array_4m)
            self.mask_16m = albula.DImage(mask_array_16m)

    def start(self, path=None, filetype=None, interval=None, stream=False):
        if filetype is not None:
            self.filetype = filetype
        if interval is not None:
            self.interval = interval
        self.stream = stream
        self.stoptimer = -1

        # HDF5
        self.eigerThread.setMonitorMode("enabled")
        if stream:
            self.eigerThread.setMonitorDiscardNew(False)
        else:
            self.eigerThread.clearMonitorBuffer()
            self.eigerThread.setMonitorDiscardNew(True)

        super().start()

    def stop(self, interval=0.0):
        if self.stoptimer < 0.0 and interval > 0.0:
            self.stoptimer = interval
            return

        print("Live view thread: Stopping thread")
        self.alive = False
        self.wait()  # waits until run stops on his own

    def check_app(self):
        try:
            self.subframe.setOffsetX(0)
        except BaseException as e:
            self.subframe = None

        if self.subframe is None:
            try:
                self.viewer.enableClose()
            except BaseException as e:
                self.viewer = None

        if self.viewer is None:
            self.viewer = albula.openMainFrame()

        if self.subframe is None:
            self.subframe = self.viewer.openSubFrame()

    def run(self):
        self.alive = True

        while self.alive:
            # hdf5
            wavelength = 12398.4 / energy

            try:
                energy = self.eigerThread.photonEnergy
                detector_distance = self.eigerThread.proxyEiger.read_attribute(
                    "DetectorDistance"
                ).value
                beam_center_x = self.eigerThread.proxyEiger.read_attribute(
                    "BeamCenterX"
                ).value
                beam_center_y = self.eigerThread.proxyEiger.read_attribute(
                    "BeamCenterY"
                ).value
            except:
                detector_distance = 1.0
                beam_center_x = 2074
                beam_center_y = 2181

            # get latest file from reveiver
            timestamp = time.time()
            if self.stream:
                try:
                    img = self.eigerMonitor.next()
                except:
                    time.sleep(0.1)
                    if self.stoptimer > 0:
                        self.stoptimer -= 0.1
                        if self.stoptimer <= 0.0:
                            self.stoptimer = -1.0
                            self.alive = False
                    continue
            else:
                try:
                    img = self.eigerMonitor.monitor()
                    self.eigerThread.clearMonitorBuffer()
                except:
                    print(sys.exc_info())
                    continue

            # display image
            img.optionalData().set_wavelength(wavelength)
            img.optionalData().set_detector_distance(detector_distance)
            img.optionalData().set_beam_center_x(beam_center_x)
            img.optionalData().set_beam_center_y(beam_center_y)
            img.optionalData().set_x_pixel_size(0.000075)
            img.optionalData().set_y_pixel_size(0.000075)

            if self.eigerMask16M is not None and img.width() == 4148:
                img.optionalData().set_pixel_mask(mask_16m)
            elif self.eigerMask4M is not None and img.width() == 2068:
                img.optionalData().set_pixel_mask(mask_4m)

            self.check_app()

            self.subframe.hideResolutionRings()
            self.subframe.loadImage(img)
            self.subframe.showResolutionRings()

            # wait interval / check if stop requested
            interval = time.time() - timestamp
            if self.stoptimer > 0:
                self.stoptimer -= interval
                if self.stoptimer <= 0.0:
                    self.stoptimer = -1.0
                    self.alive = False

            if self.stream:
                time.sleep(0.1)
                continue

            # sleep for self.interval and check stop in between
            while interval < self.interval and self.alive:
                if self.stoptimer > 0.0:
                    self.stoptimer -= 0.05
                    if self.stoptimer <= 0.0:
                        self.stoptimer = -1.0
                        self.alive = False
                time.sleep(0.05)
                interval += 0.05

        # end hdf5

        print("Live view thread: Thread for Live view died")
        self.alive = False

        if interval is not None:
            self.interval = interval


if __name__ == "__main__":
    lv = LiveView()
    lv.start()
    time.sleep(200)
    lv.stop()



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
GenericDiffractometer
"""

import copy
import enum
import json
import logging
import math
import os
import time
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import gevent
import gevent.event
import numpy
from pydantic.v1 import (
    BaseModel,
    Field,
    ValidationError,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects import sample_centring
from mxcubecore.model import queue_model_objects

__credits__ = ["MXCuBE collaboration"]

__version__ = "2.2."
__status__ = "Draft"


class DiffractometerState:
    """
    Enumeration of diffractometer states
    """

    Created = 0
    Initializing = 1
    On = 2
    Off = 3
    Closed = 4
    Open = 5
    Ready = 6
    Busy = 7
    Moving = 8
    Standby = 9
    Running = 10
    Started = 11
    Stopped = 12
    Paused = 13
    Remote = 14
    Reset = 15
    Closing = 16
    Disable = 17
    Waiting = 18
    Positioned = 19
    Starting = 20
    Loading = 21
    Unknown = 22
    Alarm = 23
    Fault = 24
    Invalid = 25
    Offline = 26

    STATE_DESC = {
        Created: "Created",
        Initializing: "Initializing",
        On: "On",
        Off: "Off",
        Closed: "Closed",
        Open: "Open",
        Ready: "Ready",
        Busy: "Busy",
        Moving: "Moving",
        Standby: "Standby",
        Running: "Running",
        Started: "Started",
        Stopped: "Stopped",
        Paused: "Paused",
        Remote: "Remote",
        Reset: "Reset",
        Closing: "Closing",
        Disable: "Disable",
        Waiting: "Waiting",
        Positioned: " Positioned",
        Starting: "Starting",
        Loading: "Loading",
        Unknown: "Unknown",
        Alarm: "Alarm",
        Fault: "Fault",
        Invalid: "Invalid",
        Offline: "Offline",
    }

    @staticmethod
    def tostring(state):
        return DiffractometerState.STATE_DESC.get(state, "Unknown")


class PhaseEnum(str, enum.Enum):
    centring = "Centring"
    data_collection = "DataCollection"
    beam_location = "BeamLocation"
    transfer = "Transfer"
    unknown = "Unknown"


class PhaseModel(BaseModel):
    value: PhaseEnum = PhaseEnum.unknown


class HeadTypeEnum(str, enum.Enum):
    no_kappa = "NO_KAPPA"
    mini_kappa = "MINI_KAPPA"
    chip = "CHIP"
    plate = "PLATE"


class HolderTypeEnum(str, enum.Enum):
    known_geometry = "KNOWN_GEOMETRY"
    free_geometry = "FREE_GEOMETRY"


class BlockShapeEnum(str, enum.Enum):
    rectangular = "RECTANGULAR"
    elliptical = "ELLIPTICAL"


class CalibrationData(BaseModel):
    top_left: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Top left corner motor position"
    )
    top_right: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Top right corner motor position"
    )
    bottom_left: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Bottom left corner motor position"
    )


class SampleHolderSectionModel(BaseModel):
    calibration_data: Optional[CalibrationData]
    section_offset: Tuple[int, int] = Field(
        [0, 0], description="Block offset in grid layout system coordinates x, y"
    )
    block_size: Tuple[float, float] = Field(
        [15, 15], description="Block size horizontal, vertical in mm"
    )
    block_spacing: Tuple[float, float] = Field(
        [15, 15], description="Spacing between blocks horizontal, vertical in mm"
    )
    block_shape: BlockShapeEnum = BlockShapeEnum.rectangular
    number_of_rows: int = Field(6, description="Numer of rows")
    number_of_collumns: int = Field(6, description="Numer of collumns")
    row_labels: List[str] = Field([], description="Row lables")
    column_lables: List[str] = Field([], description="Collumn lables")
    targets_per_block: Tuple[int, int] = Field(
        [20, 20], description="Targets per block dim1 and dim2"
    )


class ChipLayout(BaseModel):
    head_type: HeadTypeEnum = HeadTypeEnum.chip
    holder_type: HolderTypeEnum = HolderTypeEnum.known_geometry
    holder_brand: str = Field("", description="Brand/make of sample holder")
    holder_size: Tuple[float, float] = Field(
        [0, 0], description="Size of sample holder in mm horizontal and vertical"
    )
    sections: List[SampleHolderSectionModel]
    calibration_data: CalibrationData


class GonioHeadConfiguration(BaseModel):
    current: str = Field("", description="Selected chip layout")
    available: Dict[str, ChipLayout]


class GenericDiffractometer(HardwareObject):
    """
    Abstract base class for diffractometers
    """

    CENTRING_MOTORS_NAME = [
        "phi",
        "phiz",
        "phiy",
        "sampx",
        "sampy",
        "kappa",
        "kappa_phi",
        "beam_x",
        "beam_y",
    ]

    STATE_CHANGED_EVENT = "stateChanged"
    STATUS_CHANGED_EVENT = "statusChanged"
    MOTOR_POSITION_CHANGED_EVENT = "motorPositionsChanged"
    MOTOR_STATUS_CHANGED_EVENT = "motorStatusChanged"

    HEAD_TYPE_MINIKAPPA = "MiniKappa"
    HEAD_TYPE_SMARTMAGNET = "SmartMagnet"
    HEAD_TYPE_PLATE = "Plate"
    HEAD_TYPE_PERMANENT = "Permanent"

    CENTRING_METHOD_MANUAL = "Manual 3-click"
    CENTRING_METHOD_AUTO = "Computer automatic"
    CENTRING_METHOD_MOVE_TO_BEAM = "Move to beam"

    # TODO NBNB FIX THIS CONFUSION!!!
    MANUAL3CLICK_MODE = CENTRING_METHOD_MANUAL
    C3D_MODE = CENTRING_METHOD_AUTO

    PHASE_TRANSFER = "Transfer"
    PHASE_CENTRING = "Centring"
    PHASE_COLLECTION = "DataCollection"
    PHASE_BEAM = "BeamLocation"
    PHASE_UNKNOWN = "Unknown"

    def __init__(self, name):
        HardwareObject.__init__(self, name)

        # Hardware objects ----------------------------------------------------
        self.motor_hwobj_dict = {}
        self.centring_motors_list = None
        self.front_light_switch = None
        self.back_light_switch = None
        self.aperture = None
        self.beamstop = None
        self.cryo = None
        self.capillary = None
        self.use_sc = False

        # Channels and commands -----------------------------------------------
        self.channel_dict = {}
        self.used_channels_list = []
        self.command_dict = {}
        self.used_commands_list = []

        # flag for using sample_centring hwobj or not
        self.use_sample_centring = None

        # time to delay for state polling for controllers
        # not updating state inmediately after cmd started
        self.delay_state_polling = None

        self.delay_state_polling = (
            None  # time to delay for state polling for controllers
        )
        # not updating state inmediately after cmd started

        # Internal values -----------------------------------------------------
        self.ready_event = None
        self.head_type = GenericDiffractometer.HEAD_TYPE_MINIKAPPA
        self.phase_list = []
        self.grid_direction = None
        self.reversing_rotation = None

        self.beam_position = None
        self.zoom_centre = None
        self.pixels_per_mm_x = None
        self.pixels_per_mm_y = None

        self.current_state = None
        self.current_phase = None
        self.transfer_mode = None
        self.current_centring_procedure = None
        self.current_centring_method = None
        self.current_motor_positions = {}
        self.current_motor_states = {}

        self.fast_shutter_is_open = None
        self.centring_status = {"valid": False}
        self.centring_time = 0
        self.user_confirms_centring = None
        self.user_clicked_event = None
        self.automatic_centring_try_count = 1
        self.omega_reference_par = None
        self.move_to_motors_positions_task = None
        self.move_to_motors_positions_procedure = None

        self.centring_methods = {
            GenericDiffractometer.CENTRING_METHOD_MANUAL: self.start_manual_centring,
            GenericDiffractometer.CENTRING_METHOD_AUTO: self.start_automatic_centring,
            GenericDiffractometer.CENTRING_METHOD_MOVE_TO_BEAM: self.start_move_to_beam,
        }

        self.connect(self, "equipmentReady", self.equipment_ready)
        self.connect(self, "equipmentNotReady", self.equipment_not_ready)

        # HACK
        self.get_motor_positions = self.get_positions

    def init(self):
        super().init()
        # Internal values -----------------------------------------------------
        self.ready_event = gevent.event.Event()
        self.user_clicked_event = gevent.event.AsyncResult()
        self.user_confirms_centring = True

        self.beamstop = self.get_object_by_role("beamstop")
        self.aperture = self.get_object_by_role("aperture")
        self.capillary = self.get_object_by_role("capillary")
        self.cryo = self.get_object_by_role("cryo")

        # Hardware objects ----------------------------------------------------
        # if HWR.beamline.sample_view.camera is not None:
        #     self.image_height = HWR.beamline.sample_view.camera.get_height()
        #     self.image_width = HWR.beamline.sample_view.camera.get_width()
        # else:
        #     logging.getLogger("HWR").debug(
        #         "Diffractometer: " + "Camera hwobj is not defined"
        #     )

        if HWR.beamline.beam is not None:
            self.beam_position = HWR.beamline.beam.get_beam_position_on_screen()
            self.connect(
                HWR.beamline.beam, "beamPosChanged", self.beam_position_changed
            )
        else:
            self.beam_position = [0, 0]
            logging.getLogger("HWR").warning(
                "Diffractometer: " + "BeamInfo hwobj is not defined"
            )

        self.front_light_switch = self.get_object_by_role("frontlightswitch")
        self.back_light_switch = self.get_object_by_role("backlightswitch")

        # Channels -----------------------------------------------------------
        ss0 = self.get_property("used_channels")
        if ss0:
            try:
                self.used_channels_list = eval(ss0)
            except Exception:
                pass  # used the default value

        for channel_name in self.used_channels_list:
            self.channel_dict[channel_name] = self.get_channel_object(channel_name)
            if self.channel_dict[channel_name] is None:
                continue
            if channel_name == "TransferMode":
                self.connect(
                    self.channel_dict["TransferMode"],
                    "update",
                    self.transfer_mode_changed,
                )
            elif channel_name == "CurrentPhase":
                self.connect(
                    self.channel_dict["CurrentPhase"],
                    "update",
                    self.current_phase_changed,
                )
            elif channel_name == "HeadType":
                self.connect(
                    self.channel_dict["HeadType"], "update", self.head_type_changed
                )
            elif channel_name == "State":
                self.connect(self.channel_dict["State"], "update", self.state_changed)

        # Commands -----------------------------------------------------------
        try:
            self.used_commands_list = eval(self.get_property("used_commands", "[]"))
        except Exception:
            pass  # used the default value
        for command_name in self.used_commands_list:
            self.command_dict[command_name] = self.get_command_object(command_name)

        # Centring motors ----------------------------------------------------
        try:
            self.centring_motors_list = eval(self.get_property("centring_motors"))
        except Exception:
            self.centring_motors_list = GenericDiffractometer.CENTRING_MOTORS_NAME

        queue_model_objects.CentredPosition.set_diffractometer_motor_names(
            *self.centring_motors_list
        )

        for motor_name in self.centring_motors_list:
            # NBNB TODO refactor configuration, and set properties directly (see below)
            temp_motor_hwobj = self.get_object_by_role(motor_name)
            if temp_motor_hwobj is not None:
                self.motor_hwobj_dict[motor_name] = temp_motor_hwobj
                self.connect(temp_motor_hwobj, "stateChanged", self.motor_state_changed)
                self.connect(
                    temp_motor_hwobj, "valueChanged", self.centring_motor_moved
                )

                if motor_name == "phi":
                    self.connect(
                        temp_motor_hwobj,
                        "valueChanged",
                        self.emit_diffractometer_moved,
                    )
                elif motor_name == "zoom":
                    self.connect(
                        temp_motor_hwobj,
                        "predefinedPositionChanged",
                        self.zoom_motor_predefined_position_changed,
                    )
                    self.connect(
                        temp_motor_hwobj, "stateChanged", self.zoom_motor_state_changed
                    )
            else:
                logging.getLogger("HWR").warning(
                    "Diffractometer: Motor "
                    + "%s listed in the centring motor list, but not initalized"
                    % motor_name
                )

        # sample changer -----------------------------------------------------
        if HWR.beamline.sample_changer is None:
            logging.getLogger("HWR").warning(
                "Diffractometer: Sample Changer is not defined"
            )
        else:
            # By default use sample changer if it's defined and transfer_mode
            # is set to SAMPLE_CHANGER.
            # if not defined, set use_sc to True
            if self.transfer_mode is None or self.transfer_mode == "SAMPLE_CHANGER":
                self.use_sc = True

        try:
            self.use_sample_centring = self.get_property("sample_centring")
            if self.use_sample_centring:
                self.centring_phi = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["phi"], direction=-1
                )
                self.centring_phiz = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["phiz"]
                )
                self.centring_phiy = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["phiy"], direction=-1
                )
                self.centring_sampx = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["sampx"]
                )
                self.centring_sampy = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["sampy"]
                )
        except Exception:
            pass  # used the default value

        try:
            self.delay_state_polling = self.get_property("delay_state_polling")
        except Exception:
            pass

        # Other parameters ---------------------------------------------------
        try:
            self.zoom_centre = eval(self.get_property("zoom_centre"))
        except Exception:
            self.zoom_centre = {"x": 0, "y": 0}
            logging.getLogger("HWR").warning(
                "Diffractometer: " + "zoom centre not configured"
            )

        self.reversing_rotation = self.get_property("reversing_rotation")
        try:
            # grid_direction describes how a grid is collected
            # 'fast' is collection direction and 'slow' describes
            # move to the next collection line
            self.grid_direction = eval(self.get_property("grid_direction"))
        except Exception:
            self.grid_direction = {"fast": (0, 1), "slow": (1, 0), "omega_ref": 0}
            logging.getLogger("HWR").warning(
                "Diffractometer: Grid " + "direction is not defined. Using default."
            )

        try:
            self.phase_list = eval(self.get_property("phase_list"))
        except Exception:
            self.phase_list = [
                GenericDiffractometer.PHASE_TRANSFER,
                GenericDiffractometer.PHASE_CENTRING,
                GenericDiffractometer.PHASE_COLLECTION,
                GenericDiffractometer.PHASE_BEAM,
            ]

    # to make it compatibile
    def __getattr__(self, attr):
        if attr.startswith("__"):
            raise AttributeError(attr)
        else:
            if attr == "currentCentringProcedure":
                return self.current_centring_procedure
            if attr == "centringStatus":
                return self.centring_status
            if attr == "imageClicked":
                return self.image_clicked
            if attr == "cancelCentringMethod":
                return self.cancel_centring_method
            if attr == "pixelsPerMmY":
                return self.pixels_per_mm_x
            if attr == "pixelsPerMmZ":
                return self.pixels_per_mm_y
            return HardwareObject.__getattr__(self, attr)

    # Contained Objects
    # NBNB Temp[orary hack - should be cleaned up together with configuration
    @property
    def omega(self):
        """omega motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("phi")

    @property
    def centring_x(self):
        """centring_x motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("sampx")

    @property
    def centring_y(self):
        """centring_y motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("sampy")

    @property
    def alignment_x(self):
        """alignment_x motor object (also used as graphics.focus)

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("focus")

    @property
    def alignment_y(self):
        """alignment_y motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("phiy")

    @property
    def alignment_z(self):
        """alignment_z motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("phiz")

    def is_ready(self):
        """
        Detects if device is ready
        """
        return self.current_state == DiffractometerState.tostring(
            DiffractometerState.Ready
        )

    def wait_device_not_ready(self, timeout=5):
        with gevent.Timeout(timeout, Exception("Timeout waiting for device not ready")):
            while self.is_ready():
                time.sleep(0.01)

    def wait_device_ready(self, timeout=30):
        """Waits when diffractometer status is ready:

        :param timeout: timeout in second
        :type timeout: int
        """
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self.is_ready():
                time.sleep(0.01)

    wait_ready = wait_device_ready

    def execute_server_task(self, method, timeout=30, *args):
        """Method is used to execute commands and wait till
           diffractometer is in ready state

        :param method: method to be executed
        :type method: instance
        :param timeout: timeout in seconds
        :type timeout: seconds
        """
        # self.ready_event.clear()
        self.current_state = DiffractometerState.tostring(DiffractometerState.Busy)
        method(*args)
        time.sleep(5)
        # gevent.sleep(2)
        self.wait_device_ready(timeout)
        self.ready_event.set()

    def in_plate_mode(self):
        """Returns True if diffractometer in plate mod

        :returns: boolean
        """
        return self.head_type == GenericDiffractometer.HEAD_TYPE_PLATE

    def get_head_type(self):
        """Returns head type

        :returns: string
        """
        return self.head_type

    def use_sample_changer(self):
        """Returns True if sample changer is in use

        :returns: boolean
        """
        return self.use_sc

    def set_use_sc(self, flag):
        """Sets use_sc flag, that indicates if sample changer is used

        :param flag: use sample changer flag
        :type flag: boolean
        """
        if flag:
            # check both transfer_mode and sample_Changer
            if HWR.beamline.sample_changer is None:
                logging.getLogger("HWR").error(
                    "Diffractometer: Sample " + "Changer is not available"
                )
                return False

            if (
                self.transfer_mode is None
                or self.channel_dict["TransferMode"].get_value() == "SAMPLE_CHANGER"
            ):
                # if transferMode is not defined, ignore the checkup
                self.use_sc = True
            else:
                logging.getLogger("HWR").error(
                    "Diffractometer: Set the "
                    + "diffractometer TransferMode to SAMPLE_CHANGER first!!"
                )
                return False
        else:
            self.use_sc = False
        return True

    def transfer_mode_changed(self, transfer_mode):
        """
        Descript. :
        """
        logging.getLogger("HWR").info(
            "current_transfer_mode is set to %s" % transfer_mode
        )
        self.transfer_mode = transfer_mode
        if transfer_mode != "SAMPLE_CHANGER":
            self.use_sc = False
        self.emit("minidiffTransferModeChanged", (transfer_mode,))

    def get_transfer_mode(self):
        """ """
        return self.transfer_mode

    def get_current_phase(self):
        """
        Descript. :
        """
        return self.current_phase

    def get_grid_direction(self):
        """
        Descript. :
        """
        return self.grid_direction

    def get_available_centring_methods(self):
        """
        Descript. :
        """
        return self.centring_methods.keys()

    def get_current_centring_method(self):
        """
        Descript. :
        """
        return self.current_centring_method

    def is_reversing_rotation(self):
        """ """
        return self.reversing_rotation is True

    def beam_position_changed(self, value):
        """
        Descript. :
        """
        self.beam_position = list(value)

    # def get_motor_positions(self):
    #    return

    # TODO rename to get_motor_positions
    def get_positions(self):
        """
        Descript. :
        """

        self.current_motor_positions["beam_x"] = (
            self.beam_position[0] - self.zoom_centre["x"]
        ) / self.pixels_per_mm_y
        self.current_motor_positions["beam_y"] = (
            self.beam_position[1] - self.zoom_centre["y"]
        ) / self.pixels_per_mm_x

        return self.current_motor_positions

    def get_motors(self):
        """Get motor_name:Motor dictionary"""
        return self.motor_hwobj_dict.copy()

    def get_snapshot(self):
        """
        Get snapshot from sample view

        Returns:
            bytes: A bytes object of the current camera image.
        """
        if HWR.beamline.sample_view:
            return HWR.beamline.sample_view.get_snapshot()

    def save_snapshot(self, filename):
        if HWR.beamline.sample_view:
            return HWR.beamline.sample_view.save_snapshot(filename)

    def get_pixels_per_mm(self):
        """
        Returns tuple with pixels_per_mm_x and pixels_per_mm_y

        :returns: list with two floats
        """
        return (self.pixels_per_mm_x, self.pixels_per_mm_y)

    def get_phase_list(self):
        """
        Returns list of available phases

        :returns: list with str
        """
        return self.phase_list

    def start_centring_method(self, method, sample_info=None, wait=False):
        """ """

        if self.current_centring_method is not None:
            logging.getLogger("HWR").error(
                "Diffractometer: already in centring method %s"
                % self.current_centring_method
            )
            return
        curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
        self.centring_status = {
            "valid": False,
            "startTime": curr_time,
            "angleLimit": None,
        }
        self.emit_centring_started(method)

        try:
            centring_method = self.centring_methods[method]
        except KeyError as diag:
            logging.getLogger("HWR").error(
                "Diffractometer: unknown centring method (%s)" % str(diag)
            )
            self.emit_centring_failed()
        else:
            try:
                centring_method(sample_info, wait_result=wait)
            except Exception:
                logging.getLogger("HWR").exception(
                    "Diffractometer: problem while centring"
                )
                self.emit_centring_failed()

    def cancel_centring_method(self, reject=False):

        if self.current_centring_procedure is not None:
            try:
                self.current_centring_procedure.kill()
            except Exception:
                logging.getLogger("HWR").exception(
                    "Diffractometer: problem aborting the centring method"
                )
            try:
                # TODO... do we need this at all?
                # fun = self.cancel_centring_methods[self.current_centring_method]
                pass
            except KeyError:
                self.emit_centring_failed()
            else:
                try:
                    fun()
                except Exception:
                    self.emit_centring_failed()
        else:
            self.emit_centring_failed()
        self.emit_progress_message("")
        if reject:
            self.reject_centring()

    def start_manual_centring(self, sample_info=None, wait_result=None):
        self.emit_progress_message("Manual 3 click centring...")
        if self.use_sample_centring:
            self.current_centring_procedure = sample_centring.start(
                {
                    "phi": self.centring_phi,
                    "phiy": self.centring_phiy,
                    "sampx": self.centring_sampx,
                    "sampy": self.centring_sampy,
                    "phiz": self.centring_phiz,
                },
                self.pixels_per_mm_x,
                self.pixels_per_mm_y,
                self.beam_position[0],
                self.beam_position[1],
            )
        else:
            self.current_centring_procedure = gevent.spawn(self.manual_centring)
        self.current_centring_procedure.link(self.centring_done)

    def start_automatic_centring(
        self, sample_info=None, loop_only=False, wait_result=None
    ):
        self.emit_progress_message("Automatic centring...")

        while self.automatic_centring_try_count > 0:
            if self.use_sample_centring:
                self.current_centring_procedure = sample_centring.start_auto(
                    HWR.beamline.sample_view,
                    {
                        "phi": self.centring_phi,
                        "phiy": self.centring_phiy,
                        "sampx": self.centring_sampx,
                        "sampy": self.centring_sampy,
                        "phiz": self.centring_phiz,
                    },
                    self.pixels_per_mm_x,
                    self.pixels_per_mm_y,
                    self.beam_position[0],
                    self.beam_position[1],
                    msg_cb=self.emit_progress_message,
                    new_point_cb=lambda point: self.emit(
                        "newAutomaticCentringPoint", (point,)
                    ),
                )
            else:
                self.current_centring_procedure = gevent.spawn(self.automatic_centring)
            self.current_centring_procedure.link(self.centring_done)

            if wait_result:
                self.ready_event.wait()
                self.ready_event.clear()

            self.automatic_centring_try_count -= 1
        self.automatic_centring_try_count = 1

    def start_move_to_beam(
        self, coord_x=None, coord_y=None, omega=None, wait_result=None
    ):
        """
        Descript. :
        """
        try:
            self.emit_progress_message("Move to beam...")
            self.centring_time = time.time()
            curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
            self.centring_status = {
                "valid": True,
                "startTime": curr_time,
                "endTime": curr_time,
            }
            if coord_x is None and coord_y is None:
                coord_x = self.beam_position[0]
                coord_y = self.beam_position[1]

            motors = self.get_centred_point_from_coord(
                coord_x, coord_y, return_by_names=True
            )
            if omega is not None:
                motors["phi"] = omega

            self.centring_status["motors"] = motors
            self.centring_status["valid"] = True
            self.centring_status["angleLimit"] = True
            self.emit_progress_message("")
            self.accept_centring()
            self.current_centring_method = None
            self.current_centring_procedure = None
        except Exception:
            logging.exception("Diffractometer: Could not complete 2D centring")

    def centring_done(self, centring_procedure):
        """
        Descript. :
        """
        try:
            motor_pos = centring_procedure.get()
            if isinstance(motor_pos, gevent.GreenletExit):
                raise motor_pos
        except Exception:
            logging.exception("Could not complete centring")
            self.emit_centring_failed()
        else:
            self.emit_progress_message("Moving sample to centred position...")
            self.emit_centring_moving()

            try:
                logging.getLogger("HWR").debug(
                    "Centring finished. Moving motors to position %s" % str(motor_pos)
                )
                self.move_to_motors_positions(motor_pos, wait=True)
            except Exception:
                logging.exception("Could not move to centred position")
                self.emit_centring_failed()
            else:
                # if 3 click centring move -180. well. dont, in principle the calculated
                # centred positions include omega to initial position
                pass
                # if not self.in_plate_mode():
                #    logging.getLogger("HWR").debug("Centring finished. Moving omega back to initial position")
                #    self.motor_hwobj_dict['phi'].set_value_relative(-180, timeout=None)
                #    logging.getLogger("HWR").debug("         Moving omega done")

            if (
                self.current_centring_method
                == GenericDiffractometer.CENTRING_METHOD_AUTO
            ):
                self.emit("newAutomaticCentringPoint", motor_pos)
            self.ready_event.set()
            self.centring_time = time.time()
            self.emit_centring_successful()
            self.emit_progress_message("")

    def manual_centring(self):
        raise NotImplementedError

    def automatic_centring(self):
        raise NotImplementedError

    def centring_motor_moved(self, pos):
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()

    def invalidate_centring(self):
        if self.current_centring_procedure is None and self.centring_status["valid"]:
            self.centring_status = {"valid": False}
            self.emit_progress_message("")
            self.emit("centringInvalid", ())

    def emit_diffractometer_moved(self, *args):
        self.emit("diffractometerMoved", ())

    def motor_positions_to_screen(self, centred_positions_dict):
        if self.use_sample_centring:
            self.update_zoom_calibration()
            if None in (self.pixels_per_mm_x, self.pixels_per_mm_y):
                return 0, 0
            phi_angle = math.radians(
                self.centring_phi.direction * self.centring_phi.get_value()
            )
            sampx = self.centring_sampx.direction * (
                centred_positions_dict["sampx"] - self.centring_sampx.get_value()
            )
            sampy = self.centring_sampy.direction * (
                centred_positions_dict["sampy"] - self.centring_sampy.get_value()
            )
            phiy = self.centring_phiy.direction * (
                centred_positions_dict["phiy"] - self.centring_phiy.get_value()
            )
            phiz = self.centring_phiz.direction * (
                centred_positions_dict["phiz"] - self.centring_phiz.get_value()
            )
            rot_matrix = numpy.matrix(
                [
                    math.cos(phi_angle),
                    -math.sin(phi_angle),
                    math.sin(phi_angle),
                    math.cos(phi_angle),
                ]
            )
            rot_matrix.shape = (2, 2)
            inv_rot_matrix = numpy.array(rot_matrix.I)
            dx, dy = (
                numpy.dot(numpy.array([sampx, sampy]), inv_rot_matrix)
                * self.pixels_per_mm_x
            )

            x = (phiy * self.pixels_per_mm_x) + self.beam_position[0]
            y = dy + (phiz * self.pixels_per_mm_y) + self.beam_position[1]

            return x, y
        else:
            raise NotImplementedError

    def move_to_centred_position(self, centred_position):
        self.move_motors(centred_position)

    def move_to_motors_positions(self, motors_positions, wait=False):
        self.emit_progress_message("Moving to motors positions...")
        self.move_to_motors_positions_procedure = gevent.spawn(
            self.move_motors, motors_positions
        )
        self.move_to_motors_positions_procedure.link(self.move_motors_done)
        if wait:
            self.wait_device_ready(10)

    def move_motors(self, motor_positions, timeout=15):
        """
        Moves diffractometer motors to the requested positions

        :param motors_dict: dictionary with motor names or hwobj
                            and target values.
        :type motors_dict: dict
        """
        if not isinstance(motor_positions, dict):
            motor_positions = motor_positions.as_dict()

        self.wait_device_ready(timeout)

        for motor in motor_positions.keys():
            position = motor_positions[motor]
            self.log.debug(f"moving motor {motor} to position {position}")
            if isinstance(motor, str):
                motor_role = motor
                motor = self.motor_hwobj_dict.get(motor_role)
                if None in (motor, position):
                    continue
            motor.set_value(position)
        self.wait_device_ready(timeout)

        if self.delay_state_polling is not None and self.delay_state_polling > 0:
            # delay polling for state in the
            # case of controller not reporting MOVING inmediately after cmd
            gevent.sleep(self.delay_state_polling)

        self.wait_device_ready(timeout)

    def move_motors_done(self, move_motors_procedure):
        """
        Descript. :
        """
        self.move_to_motors_positions_procedure = None
        self.emit_progress_message("")

    def move_to_beam(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """
        try:
            pos = self.get_centred_point_from_coord(x, y, return_by_names=False)
            if omega is not None:
                pos["phiMotor"] = omega
            self.move_to_motors_positions(pos)
        except Exception:
            logging.getLogger("HWR").exception(
                "Diffractometer: could not center to beam, aborting"
            )

    def image_clicked(self, x, y, xi=None, yi=None):
        """
        Descript. :
        """
        if self.use_sample_centring:
            sample_centring.user_click(x, y)
        else:
            self.user_clicked_event.set((x, y))

    def accept_centring(self):
        """
        Descript. :
        Arg.      " fully_centred_point. True if 3 click centring
                    else False
        """
        self.centring_status["valid"] = True
        self.centring_status["accepted"] = True
        centring_status = self.get_centring_status()
        if "motors" not in centring_status:
            centring_status["motors"] = self.get_positions()
        self.emit("centringAccepted", (True, centring_status))
        self.emit("fsmConditionChanged", "centering_position_accepted", True)

    def reject_centring(self):
        """
        Descript. :
        """
        if self.current_centring_procedure:
            self.current_centring_procedure.kill()
        self.centring_status = {"valid": False}
        self.emit_progress_message("")
        self.emit("centringAccepted", (False, self.get_centring_status()))
        self.emit("fsmConditionChanged", "centering_position_accepted", False)

    def emit_centring_started(self, method):
        """
        Descript. :
        """
        self.current_centring_method = method
        self.emit("centringStarted", (method, False))

    def emit_centring_moving(self):
        """
        Descript. :
        """
        self.emit("centringMoving", ())

    def emit_centring_failed(self):
        """
        Descript. :
        """
        self.centring_status = {"valid": False}
        method = self.current_centring_method
        self.current_centring_method = None
        self.current_centring_procedure = None
        self.emit("centringFailed", (method, self.get_centring_status()))

    def emit_centring_successful(self):
        """
        Descript. :
        """
        if self.current_centring_procedure is not None:
            curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
            self.centring_status["endTime"] = curr_time

            motor_pos = self.current_centring_procedure.get()
            self.centring_status["motors"] = self.convert_from_obj_to_name(motor_pos)
            self.centring_status["method"] = self.current_centring_method
            self.centring_status["valid"] = True

            method = self.current_centring_method
            self.emit("centringSuccessful", (method, self.get_centring_status()))
            self.current_centring_method = None
            self.current_centring_procedure = None
        else:
            logging.getLogger("HWR").debug(
                "Diffractometer: Trying to emit "
                + "centringSuccessful outside of a centring"
            )

    def emit_progress_message(self, msg=None):
        """
        Descript. :
        """
        self.emit("progressMessage", (msg,))

    def get_centring_status(self):
        """
        Descript. :
        """
        return copy.deepcopy(self.centring_status)

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        raise NotImplementedError

    def get_point_between_two_points(
        self, point_one, point_two, frame_num, frame_total
    ):
        """
        Method returns a centring point between two centring points
        It is used to get a position on a helical line based on
        frame number and total frame number
        """
        new_point = {}
        point_one = point_one.as_dict()
        point_two = point_two.as_dict()
        for motor in point_one.keys():
            new_motor_pos = (
                frame_num
                / float(frame_total)
                * abs(point_one[motor] - point_two[motor])
                + point_one[motor]
            )
            new_motor_pos += 0.5 * (point_two[motor] - point_one[motor]) / frame_total
            new_point[motor] = new_motor_pos
        return new_point

    def convert_from_obj_to_name(self, motor_pos):
        motors = {}
        for motor_role in self.centring_motors_list:
            motor_obj = self.get_object_by_role(motor_role)
            try:
                motors[motor_role] = motor_pos[motor_obj]
            except KeyError:
                if motor_obj:
                    motors[motor_role] = motor_obj.get_value()
        motors["beam_x"] = (
            self.beam_position[0] - self.zoom_centre["x"]
        ) / self.pixels_per_mm_y
        motors["beam_y"] = (
            self.beam_position[1] - self.zoom_centre["y"]
        ) / self.pixels_per_mm_x
        return motors

    def visual_align(self, point_1, point_2):
        """
        Descript. :
        """
        return

    def move_omega_relative(self, relative_angle):
        """
        Descript. :
        """
        return

    def get_osc_limits(self):
        """Returns osc limits"""
        return

    def get_osc_max_speed(self):
        return

    def get_scan_limits(self, speed=None, num_images=None, exp_time=None):
        """
        Gets scan limits. Necessary for example in the plate mode
        where osc range is limited
        """
        return

    def set_phase(self, phase, timeout=None):
        """Sets diffractometer to selected phase
           By default available phase is Centring, BeamLocation,
           DataCollection, Transfer

        :param phase: phase
        :type phase: string
        :param timeout: timeout in sec
        :type timeout: int
        """
        if timeout:
            self.ready_event.clear()
            set_phase_task = gevent.spawn(
                self.execute_server_task,
                self.command_dict["startSetPhase"],
                timeout,
                phase,
            )
            self.ready_event.wait()
            self.ready_event.clear()
        else:
            self.command_dict["startSetPhase"](phase)

    def update_zoom_calibration(self):
        self.pixels_per_mm_x = 1.0 / self.channel_dict["CoaxCamScaleX"].get_value()
        self.pixels_per_mm_y = 1.0 / self.channel_dict["CoaxCamScaleY"].get_value()
        self.emit("pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y)))

    def zoom_motor_state_changed(self, state):
        self.emit("zoomMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def zoom_motor_predefined_position_changed(self, position_name, offset):
        self.update_zoom_calibration()
        self.emit("zoomMotorPredefinedPositionChanged", (position_name, offset))

    def equipment_ready(self):
        self.emit("minidiffReady", ())

    def equipment_not_ready(self):
        self.emit("minidiffNotReady", ())

    def motor_state_changed(self, state):
        """
        Descript. :
        """
        self.emit("minidiffStateChanged", (state,))

    def current_phase_changed(self, current_phase):
        """
        Descript. :
        """
        self.current_phase = current_phase
        if current_phase != GenericDiffractometer.PHASE_UNKNOWN:
            logging.getLogger("GUI").info(
                "Diffractometer: Current phase changed to %s" % current_phase
            )
        self.emit("minidiffPhaseChanged", (current_phase,))

    def sample_is_loaded_changed(self, sample_is_loaded):
        """
        Descript. :
        """
        self.sample_is_loaded = sample_is_loaded
        self.emit("minidiffSampleIsLoadedChanged", (sample_is_loaded,))

    def head_type_changed(self, head_type):
        """
        Descript. :
        """
        self.head_type = head_type
        self.emit("minidiffHeadTypeChanged", (head_type,))

        if "SampleIsLoaded" not in str(self.used_channels_list):
            return
        try:
            self.disconnect(
                self.channel_dict["SampleIsLoaded"],
                "update",
                self.sample_is_loaded_changed,
            )
        except Exception:
            pass

        if (
            head_type == GenericDiffractometer.HEAD_TYPE_MINIKAPPA
            or head_type == GenericDiffractometer.HEAD_TYPE_SMARTMAGNET
        ):
            self.connect(
                self.channel_dict["SampleIsLoaded"],
                "update",
                self.sample_is_loaded_changed,
            )
        else:
            logging.getLogger("HWR").info(
                "Diffractometer: SmartMagnet "
                + "is not available, only works for Minikappa and SmartMagnet head"
            )

    def move_kappa_and_phi(self, kappa, kappa_phi):
        return

    def close_kappa(self):
        """
        Descript. :
        """
        return

    def get_osc_dynamic_limits(self):
        return (-10000, 10000)

    def zoom_in(self):
        return

    def zoom_out(self):
        return

    def save_centring_positions(self):
        pass

    def force_emit_signals(self):
        for motor_hwobj in self.motor_hwobj_dict.values():
            motor_hwobj.force_emit_signals()

    def get_head_configuration(self) -> Union[GonioHeadConfiguration, None]:
        chip_def_fpath = self.get_property("chip_definition_file", "")
        chip_def_fpath = HWR.get_hardware_repository().find_in_repository(
            chip_def_fpath
        )

        data = None

        if os.path.isfile(chip_def_fpath):
            with open(chip_def_fpath, "r") as _f:
                chip_def = json.load(_f)

                try:
                    data = GonioHeadConfiguration(**chip_def)
                except ValidationError:
                    logging.getLogger("HWR").exception(
                        "Validation error in %s" % chip_def_fpath
                    )

        return data

    def set_head_configuration(self, str_data: str) -> None:
        data = json.loads(str_data)

        chip_def_fpath = self.get_property("chip_definition_file", "")
        chip_def_fpath = HWR.get_hardware_repository().find_in_repository(
            chip_def_fpath
        )

        if os.path.isfile(chip_def_fpath):
            with open(chip_def_fpath, "w+") as _f:
                try:
                    GonioHeadConfiguration(**data)
                except ValidationError:
                    logging.getLogger("HWR").exception(
                        "Validation error in %s" % chip_def_fpath
                    )
                else:
                    _f.write(json.dumps(data, indent=4))

    def set_chip_layout(self, layout_name: str) -> bool:
        data = self.get_head_configuration().dict()
        data["current"] = layout_name
        self.set_head_configuration(json.dumps(data))
        return True



import json

from pydantic.v1 import (
    BaseModel,
    Field,
)

from mxcubecore.model.common import (
    CommonCollectionParamters,
    LegacyParameters,
    PathParameters,
    StandardCollectionParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry, TaskPrerequisite

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class TestUserCollectionParameters(BaseModel):
    num_images: int = Field(0, description="")
    exp_time: float = Field(100e-6, gt=0, lt=1, description="s")

    class Config:
        extra: "ignore"


class TestCollectionTaskParameters(BaseModel):
    path_parameters: PathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    user_collection_parameters: TestUserCollectionParameters
    legacy_parameters: LegacyParameters

    @staticmethod
    def update_dependent_fields(field_data):
        return {}

    @staticmethod
    def ui_schema():
        return json.dumps({})


class TestCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class TestCollectionQueueEntry(BaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = TestCollectionQueueModel
    DATA_MODEL = TestCollectionTaskParameters
    NAME = "TestCollection"
    REQUIRES = [
        TaskPrerequisite.POINT,
        TaskPrerequisite.LINE,
        TaskPrerequisite.CHIP,
        TaskPrerequisite.MESH,
        TaskPrerequisite.NO_SHAPE_2D,
    ]

    def __init__(self, view, data_model: TestCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()



import enum
import logging
import time

from gevent import Timeout

from mxcubecore.BaseHardwareObjects import (
    HardwareObjectState,
)
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

"""
Interfaces Sardana Motor objects.
taurusname is the only obligatory property.

.. code-block:: xml

   <object class="SardanaMotor">
    <taurusname>dmot01</taurusname>
    <username>Dummy</username>
    <actuator_name>dummy_motor</actuator_name>
    <threshold>0.005</threshold>
    <move_threshold>0.005</move_threshold>
    <interval>2000</interval>
   </object>
"""


class SardanaMotorState(enum.Enum):
    READY = HardwareObjectState.READY
    ON = HardwareObjectState.READY
    OFF = HardwareObjectState.OFF
    MOVING = HardwareObjectState.BUSY
    STANDBY = HardwareObjectState.READY
    FAULT = HardwareObjectState.FAULT
    INIT = HardwareObjectState.BUSY
    RUNNING = HardwareObjectState.BUSY
    ALARM = HardwareObjectState.WARNING
    DISABLE = HardwareObjectState.OFF
    UNKNOWN = HardwareObjectState.UNKNOWN
    INVALID = HardwareObjectState.FAULT


class SardanaMotor(AbstractMotor):
    suffix_position = "Position"
    suffix_state = "State"
    suffix_stop = "Stop"
    suffix_velocity = "Velocity"
    suffix_acceleration = "Acceleration"

    def __init__(self, name):
        super().__init__(name)
        self.stop_command = None
        self.position_channel = None
        self.state_channel = None
        self.taurusname = None
        self.motor_position = 0.0
        self.threshold_default = 0.0018
        self.move_threshold_default = 0.0
        self.polling_default = "events"
        self.limit_upper = None
        self.limit_lower = None
        self.static_limits = (-1e4, 1e4)
        self.limits = (None, None)

    def init(self):
        super().init()

        self.taurusname = self.get_property("taurusname")
        if not self.taurusname:
            raise RuntimeError("Undefined property taurusname")

        self.actuator_name = self.get_property("actuator_name")
        if not self.name:
            logging.getLogger("HWR").info(
                "Undefined property actuator_name in xml. Applying name during instance creation."
            )
            self.actuator_name = self.name

        self.threshold = self.get_property("threshold", self.threshold_default)
        logging.getLogger("HWR").debug(
            "Motor {0} threshold = {1}".format(self.actuator_name, self.threshold)
        )

        self.move_threshold = self.get_property(
            "move_threshold", self.move_threshold_default
        )
        logging.getLogger("HWR").debug(
            "Motor {0} move_threshold = {1}".format(
                self.actuator_name, self.move_threshold
            )
        )

        self.polling = self.get_property("interval", self.polling_default)
        logging.getLogger("HWR").debug(
            "Motor {0} polling = {1}".format(self.actuator_name, self.polling)
        )

        self.stop_command = self.add_command(
            {
                "type": "sardana",
                "name": self.actuator_name + SardanaMotor.suffix_stop,
                "taurusname": self.taurusname,
            },
            "Stop",
        )
        self.position_channel = self.add_channel(
            {
                "type": "sardana",
                "name": self.actuator_name + SardanaMotor.suffix_position,
                "taurusname": self.taurusname,
                "polling": self.polling,
            },
            "Position",
        )
        self.state_channel = self.add_channel(
            {
                "type": "sardana",
                "name": self.actuator_name + SardanaMotor.suffix_state,
                "taurusname": self.taurusname,
                "polling": self.polling,
            },
            "State",
        )

        self.velocity_channel = self.add_channel(
            {
                "type": "sardana",
                "name": self.actuator_name + SardanaMotor.suffix_velocity,
                "taurusname": self.taurusname,
            },
            "Velocity",
        )

        self.acceleration_channel = self.add_channel(
            {
                "type": "sardana",
                "name": self.actuator_name + SardanaMotor.suffix_acceleration,
                "taurusname": self.taurusname,
            },
            "Acceleration",
        )

        self.position_channel.connect_signal("update", self.update_value)
        self.state_channel.connect_signal("update", self._update_state)

        self.limits = self.get_limits()
        self.update_state()
        self.update_value()

    def get_state(self) -> HardwareObjectState:
        """Get the motor state

        Returns:
            Motor state
        """
        try:
            _state = self.state_channel.get_value()
            self.specific_state = _state
            return SardanaMotorState[_state.name].value
        except (KeyError, AttributeError):
            return self.STATES.UNKNOWN

    def _update_state(self, state):
        try:
            state = state.upper()
            state = SardanaMotorState[state].value
        except (AttributeError, KeyError):
            state = self.STATES.UNKNOWN
        return self.update_state(state)

    def is_ready(self) -> bool:
        """
        Returns:
            True if the motor is ready
        """
        return self.get_state() == HardwareObjectState.READY

    def wait_ready(self, timeout=None):
        with Timeout(timeout, RuntimeError("Timeout waiting for status ready")):
            while not self.is_ready():
                time.sleep(0.1)

    def is_moving(self) -> bool:
        """
        Returns:
            True if the motor is currently moving
        """
        return self.get_state() == HardwareObjectState.BUSY

    def wait_end_of_move(self, timeout: float = None) -> None:
        """
        Wait till the motor stops
        """
        with Timeout(timeout):
            # Wait a bit to ensure the motor started moving
            # 0.1 empirically obtained
            time.sleep(0.1)
            while self.is_moving():
                time.sleep(0.1)

    def get_limits(self):
        """
        Descript. : returns motor limits. If no limits channel defined then
                    static_limits is returned
        """
        try:
            self._nominal_limits = (
                self.position_channel.info.minval,
                self.position_channel.info.maxval,
            )
            return self._nominal_limits
        except Exception:
            return (None, None)

    def get_value(self):
        """
        Descript. : returns the current position
        """
        self.motor_position = self.position_channel.get_value()
        return self.motor_position

    def _set_value(self, value):
        """
        Descript. : move to the given position
        """
        self.position_channel.set_value(value)

    def stop(self):
        """
        Descript. : stops the motor immediately
        """
        self.stop_command()

    def get_velocity(self):
        try:
            return self.velocity_channel.get_value()
        except Exception:
            return None

    def set_velocity(self, value):
        self.velocity_channel.set_value(value)

    def get_acceleration(self):
        try:
            return self.acceleration_channel.get_value()
        except Exception:
            return None



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Microdiff with Exporter implementation of AbstartNState
Example xml file:
<object class="ExporterNState">
  <username>Fluorescence Detector</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <value_channel_name>FluoDetectorIsBack</value_channel_name>
  <state_channel_name>State</state_channel_name>
  <values>{"IN": False, "OUT": True}</values>
  <value_state>True</value_state>
</object>
"""

from enum import Enum

from gevent import (
    Timeout,
    sleep,
)

from mxcubecore.Command.Exporter import Exporter
from mxcubecore.Command.exporter.ExporterStates import ExporterStates
from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState

__copyright__ = """ Copyright © 2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class ExporterNState(AbstractNState):
    """Microdiff with Exporter implementation of AbstartNState"""

    SPECIFIC_STATES = ExporterStates

    def __init__(self, name):
        super().__init__(name)
        self._exporter = None
        self.value_channel = None
        self.state_channel = None
        self.use_value_as_state = None

    def init(self):
        """Initialise the device"""
        super().init()
        value_channel = self.get_property("value_channel_name")
        # use the value to check if action finished.
        self.use_value_as_state = self.get_property("value_state")
        state_channel = self.get_property("state_channel_name", "State")

        _exporter_address = self.get_property("exporter_address")
        _host, _port = _exporter_address.split(":")
        self._exporter = Exporter(_host, int(_port))

        self.value_channel = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": _exporter_address,
                "name": value_channel.lower(),
            },
            value_channel,
        )
        self.value_channel.connect_signal("update", self._update_value)

        self.state_channel = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": _exporter_address,
                "name": "state",
            },
            state_channel,
        )

        self.state_channel.connect_signal("update", self._update_state)
        self.update_state()

    def _wait_hardware(self, value, timeout=None):
        """Wait timeout seconds till hardware in place.
        Args:
            value (str, int): value to be tested.
            timeout(float): Timeout [s]. None means infinite timeout.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting for hardware")):
            while self.value_channel.get_value() != value:
                sleep(0.5)

    def _wait_ready(self, timeout=None):
        """Wait timeout seconds till status is ready.
        Args:
            timeout(float): Timeout [s]. None means infinite timeout.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting for status ready")):
            while not self.get_state() == self.STATES.READY:
                sleep(0.5)

    def _update_value(self, value):
        super().update_value(self.value_to_enum(value))

    def _update_state(self, state=None):
        """To be used to update the state when emiting the "update" signal.
        Args:
            state (str): optional state value
        Returns:
            (enum 'HardwareObjectState'): state.
        """
        if not state:
            state = self.get_state()
        else:
            state = self._str2state(state)

        return self.update_state(state)

    def _str2state(self, state):
        """Convert string state to HardwareObjectState enum value.
        Args:
            state (str): the state
        Returns:
            (enum 'HardwareObjectState'): state
        """
        try:
            return self.SPECIFIC_STATES.__members__[state.upper()].value
        except (AttributeError, KeyError):
            return self.STATES.UNKNOWN

    def get_state(self):
        """Get the device state.
        Returns:
            (enum 'HardwareObjectState'): Device state.
        """
        state = self.state_channel.get_value()
        return self._str2state(state)

    def abort(self):
        """Stop the action."""
        if self.get_state() != self.STATES.UNKNOWN:
            self._exporter.execute("abort")

    def _set_value(self, value):
        """Set device to value
        Args:
            value (str, int, float or enum): Value to be set.
        """
        # NB Workaround beacuse diffractomer does not send event on
        # change of actuators (light, scintillator, cryostream...)
        self.update_state(self.STATES.BUSY)

        if isinstance(value, Enum):
            if isinstance(value.value, (tuple, list)):
                value = value.value[0]
            else:
                value = value.value
        self.value_channel.set_value(value)
        # wait until the hardware returns value set
        if self.use_value_as_state:
            self._wait_hardware(value, 120)
        self._wait_ready(120)
        self.update_state(self.STATES.READY)

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): Enum member, corresponding to the value or UNKNOWN.
        """
        _val = self.value_channel.get_value()
        return self.value_to_enum(_val)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract beam hardware object."""

from __future__ import annotations

__copyright__ = """ Copyright © by MXCuBE Collaboration """
__license__ = "LGPLv3+"


import abc
import sys
from enum import (
    Enum,
    unique,
)
from warnings import warn

from gevent import (
    Timeout,
    sleep,
)

from mxcubecore.BaseHardwareObjects import HardwareObject


@unique
class BeamShape(Enum):
    """Beam shape definitions"""

    UNKNOWN = "unknown"
    RECTANGULAR = "rectangular"
    ELLIPTICAL = "ellipse"


class AbstractBeam(HardwareObject):
    """Abstract beam hardware object.

    Has methods to define the size and shape of the beam.

    Emits:
        beamSizeChanged(tuple[float, float]):
            Two-item tuple of beam width and beam height in micrometers
            emitted when the beam size has changed.
        beamInfoChanged(dict):
            Dictionary containing beam info emitted when the beam info has changed.

    Attributes:
        _aperture: reference to the aperture hardware object
        _slits: reference to the slits hardware object
        _definer: reference to the slits hardware object
        _beam_size_dict (dict): dictionary containing min max of aperure,
                                slits and definer.
        _beam_width (float): beam size in horizontal direction
        _beam_height (float): beam size in vertical direction
        _beam_shape (str): beam shape (rectangular, ellipse, unknown)
        _beam_divergence (tuple): beam divergence in horizontal and vertical directions
        _beam_position_on_screen (list): beam position in pixel units
        _beam_info_dict (dict): dictionary containing size_x, size_y, shape, label
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, name) -> None:
        super().__init__(name)

        self._aperture = None
        self._slits = None
        self._definer = None
        self._definer_type = None

        self._beam_size_dict = {
            "aperture": [sys.float_info.max, sys.float_info.max],
            "slits": [sys.float_info.max, sys.float_info.max],
            "definer": [sys.float_info.max, sys.float_info.max],
        }
        self._beam_width = None
        self._beam_height = None
        self._beam_shape = BeamShape.UNKNOWN
        self._beam_label = None
        self._beam_divergence = (None, None)
        self._beam_position_on_screen = [None, None]

        self._beam_info_dict = {
            "size_x": self._beam_width,
            "size_y": self._beam_height,
            "shape": self._beam_shape,
            "label": self._beam_label,
        }

    def init(self) -> None:
        """Initialise default values and objects."""
        super().init()
        _divergence_vertical = self.get_property("beam_divergence_vertical")
        _divergence_horizontal = self.get_property("beam_divergence_horizontal")
        self._beam_divergence = (_divergence_horizontal, _divergence_vertical)
        self._beam_position_on_screen = [0, 0]
        self._definer_type = self.get_property("definer_type")

    @property
    def aperture(self):
        """Aperture hardware object."""
        return self._aperture

    @property
    def slits(self):
        """Slits hardware object."""
        return self._slits

    @property
    def definer(self):
        """Beam definer hardware object."""
        return self._definer

    def get_beam_divergence(self) -> tuple:
        """Get the beam divergence.

        Returns:
            Beam divergence (horizontal, vertical) in micrometers.
        """
        return self._beam_divergence

    def get_available_size(self) -> dict:
        """Get the available beam definers configuration.

        Returns:
            Dictionary ``{"type": (list), "values": (list)}`` where
            ``type`` is the definer type ``("aperture", "slits","definer")`` and
            ``values`` is a list of available beam size definitions
            according to ``type``.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def get_defined_beam_size(self) -> dict:
        """Get the predefined beam labels and size.

        Returns:
            Dictionary with list of available beam size labels
            and the corresponding size (width,height) tuples.
            ``{"label": [str, str, ...], "size": [(w,h), (w,h), ...]}``.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def get_beam_shape(self) -> BeamShape:
        """Get the beam shape.

        Returns:
            Beam shape.
        """
        self.evaluate_beam_info()
        return self._beam_shape

    def get_beam_size(self) -> tuple[float, float]:
        """Get yhe beam size.

        Returns:
            Two-item tuple: width and height.
        """
        self.evaluate_beam_info()
        return self._beam_width, self._beam_height

    def get_value(self) -> tuple[float, float, BeamShape, str]:
        """Get the beam size (width and heigth), shape and label.
           The size is in milimeters.

        Retunrs:
            Four-item tuple: width, heigth, shape, name
        """
        return self._get_value()

    def _get_value(self) -> tuple[float, float, BeamShape, str]:
        """Implement specific get_value."""

    def set_value(self, size: list[float] | str | None = None) -> None:
        """Set the beam size.
        Args:
            size: List of width and heigth in micrometers or
                aperture or definer name as string.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def set_beam_size_shape(
        self,
        beam_width: float,
        beam_height: float,
        beam_shape: BeamShape,
    ) -> None:
        """Set beam size and shape.

        Args:
            beam_width: Requested beam width in microns.
            beam_height: Requested beam height in microns.
            beam_shape: Requested beam shape.
        """
        warn(
            "set_beam_size_shape is deprecated. Use set_value() instead",
            DeprecationWarning,
        )

        if beam_shape == BeamShape.RECTANGULAR:
            self._slits.set_horizontal_gap(beam_width)
            self._slits.set_vertical_gap(beam_height)
        elif beam_shape == BeamShape.ELLIPTICAL:
            self._aperture.set_diameter_size(beam_width)

    def get_beam_position_on_screen(self) -> list[int, int]:
        """Get the beam position.

        Returns:
            X and Y coordinates of the beam position in pixels.
        """
        # (TODO) move this method to AbstractSampleView
        return self._beam_position_on_screen

    def set_beam_position_on_screen(self, beam_x_y: list) -> None:
        """Set the beam position.

        Args:
            beam_x_y: X and Y coordinates of the beam position in pixels.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def get_beam_info_dict(self) -> dict:
        """Get beam info dictionary.

        Returns:
            Copy of beam info dictionary.
        """
        return self._beam_info_dict.copy()

    def evaluate_beam_info(self) -> dict:
        """Method called if aperture, slits or focusing has been changed.

        Evaluates which of the beam size defining devices determins the size.

        Returns:
            Beam info dictionary ``dict``, type of the definer ``str``.
            ``{size_x: float, size_y: float, shape: BeamShape enum, label: str}``.
        """
        _shape = BeamShape.UNKNOWN
        _size = min(self._beam_size_dict.values())
        key = [k for k, v in self._beam_size_dict.items() if v == _size]

        if len(key) == 1:
            _label = key[0]
        elif self._definer_type in key:
            _label = self._definer_type
        else:
            _label = "UNKNOWN"

        _shape = BeamShape.RECTANGULAR if _label == "slits" else BeamShape.ELLIPTICAL

        self._beam_width = _size[0]
        self._beam_height = _size[1]
        self._beam_shape = _shape
        self._beam_info_dict["size_x"] = _size[0]
        self._beam_info_dict["size_y"] = _size[1]
        self._beam_info_dict["shape"] = _shape
        self._beam_info_dict["label"] = _label

        return self._beam_info_dict

    def re_emit_values(self):
        """Reemit ``beamSizeChanged`` and ``beamInfoChanged`` signals."""

        HardwareObject.re_emit_values(self)
        if self._beam_width != 9999 and self._beam_height != 9999:
            self.emit("beamSizeChanged", (self._beam_width, self._beam_height))
            self.emit("beamInfoChanged", (self._beam_info_dict))
            self.emit("beamPosChanged", (self._beam_position_on_screen,))

    @property
    def is_beam(self) -> bool:
        """Check if there is beam.

        Returns:
            ``True`` if beam present, ``False`` otherwise.
        """
        return self._is_beam()

    def _is_beam(self) -> bool:
        """Specific implementation to check the presence of the beam."""
        return True

    def wait_for_beam(self, timeout: float | None = None):
        """Wait until beam present.

        Args:
            Optional - timeout in seconds,
            If timeout == 0: return at once and do not wait.
            if timeout is None: wait forever (default).

        Raises:
            RuntileError if no beam after the specified timeout.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting for beam")):
            while not self.is_beam():
                sleep(0.5)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
TangoMotor class defines motor in the Tango control system (used and tested in DESY/P11
"""

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__credits__ = ["DESY P11"]
__license__ = "LGPLv3+"
__category__ = "Motor"


class TangoMotor(AbstractMotor):
    """TangoMotor class defines motor in the Tango control system"""

    default_polling = 500

    def __init__(self, name):
        AbstractMotor.__init__(self, name)

        self.chan_position = None
        self.chan_state = None
        self.chan_limits = None
        self.cmd_set_position = None
        self.cmd_stop_axis = None
        self.cmd_set_online = None
        self.latest_position = None
        self.auto_on = False
        self.cmd_on = None
        self.cmd_calibrate = None

        self.step_limits = None

    def init(self):
        """Connects to all Tango channels and commands"""
        self.polling = self.get_property("polling", TangoMotor.default_polling)
        self.actuator_name = self.get_property("actuator_name", self.name())
        self._tolerance = self.get_property("tolerance", 1e-3)

        self.is_simulation = self.get_property("simulation",False)
        self.auto_on = self.get_property("auto_on",False)
        if self.auto_on:
            self.log.debug("AUTO_ON is set for motor %s" % self.name())

        if self.is_simulation:
            self.simulated_pos = 0.0
            self.set_ready()
            return

        self.chan_position = self.get_channel_object("axisPosition", optional=True)
        if self.chan_position is None:
            self.chan_position = self.add_channel(
                  {
                      "type": "tango",
                      "name": "axisPosition",
                      "tangoname": self.tangoname,
                      "polling": self.polling,
                  }, "Position",)

        if self.chan_position is not None:
            self.chan_position.connect_signal("update", self.update_value)
            self.update_value(self.chan_position.get_value())

        self.chan_state = self.get_channel_object("axisState", None)
        if self.chan_state is None:
            self.chan_state = self.add_channel(
                {
                    "type": "tango",
                    "name": "axisState",
                    "tangoname": self.tangoname,
                    "polling": self.polling,
                },
                "State",
            )

        if self.chan_state is not None:
            self.chan_state.connect_signal("update", self.motor_state_changed)

        self.chan_limits = self.get_channel_object("axisLimits", optional=True)
        if self.chan_limits is not None:
            self.chan_limits.connect_signal("update", self.update_limits)
            self.update_limits(self.chan_limits.get_value())
        else:
            try:
                if self.get_property("default_limits"):
                    self.update_limits(eval(self.get_property("default_limits")))
            except Exception:
                pass

        self.cmd_stop = self.get_command_object("stopAxis")
        if self.cmd_stop is None:
             self.cmd_stop = self.add_command(
                  {
                      "type": "tango",
                      "name": "stopAxis",
                      "tangoname": self.tangoname,
                  }, "Stop",)

        self.cmd_on = self.get_command_object("onCmd")
        if self.cmd_on is None:
             self.cmd_on = self.add_command(
                  {
                      "type": "tango",
                      "name": "onCmd",
                      "tangoname": self.tangoname,
                  }, "On",)


        self.chan_velocity = self.get_channel_object("velocity", optional=True)

        self.cmd_calibrate = self.get_command_object("calibrate")

        # update values
        self.motor_state_changed()
        self.update_value()

    def connectNotify(self, signal):
        """
        :param signal: signal
        :type signal: signal
        """
        if signal == "stateChanged":
            self.motor_state_changed()
        elif signal == "limitsChanged":
            self.update_limits()
        elif signal == "valueChanged":
            self.update_value()

    def get_value(self):
        if self.is_simulation:
             return self.simulated_pos
        value = self.chan_position.get_value()
        return value

    def get_velocity(self):
        if self.chan_velocity is not None:
            return self.chan_velocity.get_value()

    def set_velocity(self, value):
        if self.chan_velocity is not None:
            self.chan_velocity.set_value(value)

    def motstate_to_state(self, motstate):

        motstate = str(motstate)

        if motstate == "ON":
            state = self.STATES.READY
        elif motstate == "MOVING":
            state = self.STATES.BUSY
        elif motstate == "FAULT":
            state = self.STATES.FAULT
        elif motstate == "OFF":
            state = self.STATES.OFF
        else:
            state = self.STATES.UNKNOWN

        return state

    def motor_state_changed(self, state=None):
        if state is None:
            state = self.chan_state.get_value()

        self.update_state(self.motstate_to_state(state))

    def set_ready(self):
        self.update_state(self.STATES.READY)

    def is_moving(self):
        return ( (self.get_state() == self.STATES.BUSY ) or (self.get_state() == self.SPECIFIC_STATES.MOVING))

    def abort(self):
        """Stops motor movement"""
        self.cmd_stop()

    def calibrate(self, value):
        self.cmd_calibrate(value)

    def _set_value(self, value):
        """
        Main move method
        :param value: float
        :return:
        """
        self.log.debug("TangoMotor.py - Moving motor %s to %s" % (self.id, value))
        if self.is_simulation:
             self.simulated_pos = value
        else:
             self.start_moving()
             self.chan_position.set_value(value)

    def start_moving(self):
        self.motor_state_changed("MOVING")

        if self.auto_on:
            state = str(self.chan_state.get_value())
            if state == "OFF":
                self.cmd_on()
        # ensure that the state is updated at least once after the polling time
        # in case we miss the state update
        gevent.spawn(self._update_state)

    def _update_state(self):
        gevent.sleep(0.5)
        motor_state = self.chan_state.get_value()
        self.log.debug(" reading motor state for %s is %s" % (self.id, str(motor_state)))
        self.motor_state_changed(motor_state)

    def update_value(self, value=None):
        """Updates motor position"""
        if value is None:
            value = self.get_value()
        self.latest_value = value
        super(TangoMotor, self).update_value(value)

    def get_motor_mnemonic(self):
        """
        Returns motor mnemonic
        :return:
        """
        return "TangoMotor"



from bliss.config import static

from mxcubecore.BaseHardwareObjects import HardwareObject


class BlissVolpi(HardwareObject):
    def __init__(self, name):
        # AbstractMotor.__init__(self, name)
        super().__init__(name)

    def init(self):
        self.username = self.volpi_name

        cfg = static.get_config()
        self.volpi = cfg.get(self.volpi_name)
        self.connect(self.volpi, "intensity", self.intensity_changed)

    def connect_notify(self, signal):
        if signal == "intensityChanged":
            self.emit("intensityChanged", (self.get_value(),))

    def set_value(self, intensity):
        """set volpi to new value."""
        self.volpi.intensity = intensity

    def get_value(self):
        """get volpi intensity value."""
        return self.volpi.intensity

    def re_emit_values(self):
        self.emit("intensityChanged", (self.get_value(),))

    def intensity_changed(self, new_intensity):
        self.emit("intensityChanged", (new_intensity,))



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
ID23-2 Beam Definer.

Example xml configuration:

.. code-block:: xml

 <object class="ESRF.ESRFBeamDefiner"
   <username>Beam Definer</username>
   <object href="/bliss" role="controller"/>
   # bliss tranfocator object names
   <tf1>tf</tf1>
   <tf2>tf2</tf2>
   <beam_config>
      <name>4x4 um</name>
      <beam_size>0.004, 0.004</beam_size>
      <tf1>0 0 0 0 0 0 0 0 0</tf1>
      <tf2>0 0 1 1 0 0 0 0 0</tf2>
   </beam_config>
   <beam_config>
      <name>4x8 um</name>
      <beam_size>0.004, 0.008</beam_size>
      <tf1>0 0 0 0 0 0 0 0 0</tf1>
      <tf2>0 0 1 0 1 0 0 0 0</tf2>
   </beam_config>
   <default_size_name>4x4 um</default_size_name>
 </object>
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


from enum import Enum

from gevent import (
    Timeout,
    sleep,
)

from mxcubecore.HardwareObjects.ESRF.ESRFBeamDefiner import ESRFBeamDefiner


class ID232BeamDefiner(ESRFBeamDefiner):
    """ID23-2 beam definer implementattion"""

    def __init__(self, *args):
        super().__init__(*args)
        self.tf_cfg = {}
        self.controller = None
        self.tf1 = None
        self.tf2 = None

    def init(self):
        """Initialisation"""
        super().init()

        self.controller = self.get_object_by_role("controller")
        self.tf1 = self.controller.config.get(self.get_property("tf1"))
        self.tf2 = self.controller.config.get(self.get_property("tf2"))
        for beam_cfg in self.config:
            name = beam_cfg.get_property("name")
            tf1 = [int(x) for x in beam_cfg.get_property("tf1").split()]
            tf2 = [int(x) for x in beam_cfg.get_property("tf2").split()]
            self.tf_cfg[name] = {"tf1": tf1, "tf2": tf2}

        self.connect(self.tf1, "state", self._tf_update_state)
        self.connect(self.tf2, "state", self._tf_update_state)

        if self.get_value() == self.VALUES.UNKNOWN and self._default_name:
            # set default beam value
            self.set_value(self._default_name)

    def get_state(self):
        """Get the device state.
        Returns:
            (enum 'HardwareObjectState'): Device state.
        """
        return self.STATES.READY

    def _tf_update_state(self, state=None):
        """Update the value"""
        name = self.get_current_position_name()
        self.emit("valueChanged", name)

    def get_current_status(self):
        """Get the status of the transfocators.
        Returns:
            (tuple): Tuple of two lists, giving the state for each lense
        """
        tf1_status = self.tf1.wago.get("stat")[::2]
        tf2_status = self.tf2.wago.get("stat")[::2]
        return tf1_status, tf2_status

    def get_current_position_name(self):
        """Get the current beam size name.
        Returns:
            (str): Current beam size name.
        """
        try:
            tf1_state, tf2_state = self.get_current_status()
        except ValueError:
            return "UNKNOWN"

        for name in self.beam_config:
            if (
                self.tf_cfg[name]["tf1"] == tf1_state
                and self.tf_cfg[name]["tf2"] == tf2_state
            ):
                return name
        return "UNKNOWN"

    def set_value(self, value, timeout=None):
        """Set the beam size.
        Args:
            value(str): name of the beam size to set.
            timeout(float): Timeout to wait for the execution to finish [s].
        Raises:
            RuntimeError: Cannot change beam size.
        """
        if isinstance(value, Enum):
            value = value.name

        tf1_cfg  = self.tf_cfg[value]["tf1"]
        tf2_cfg = self.tf_cfg[value]["tf2"]
        self.tf1.set(*tf1_cfg)
        self.tf2.set(*tf2_cfg)

        try:
            self.wait_status((tf1_cfg, tf2_cfg), timeout)
        except RuntimeError as err:
            raise RuntimeError("Cannot change beam size") from err

    def wait_status(self, status, timeout=None):
        """Wait timeout seconds until status reached
        Args:
            status (tuple): Transfocator status to be reached.
            timeout (float): Timeout [s]. Defaults to None.
        """
        with Timeout(timeout, RuntimeError("Execution timeout")):
            while status != self.get_current_status():
                sleep(0.5)



from bliss.config import static

from mxcubecore.BaseHardwareObjects import HardwareObject


class BlissTurret(HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.username = self.turret_name

        cfg = static.get_config()
        self.turret = cfg.get(self.turret_name)
        self.connect(self.turret, "position", self.position_changed)
        self.connect(self.turret, "mode", self.mode_changed)

    def connect_notify(self, signal):
        if signal == "positionChanged":
            self.emit("positionChanged", (self.get_value(),))
        elif signal == "modeChanged":
            self.emit("modeChanged", (self.get_mode(),))

    def position_changed(self, new_position):
        # print self.name(), absolutePosition
        self.emit("positionChanged", (new_position,))

    def mode_changed(self, new_mode):
        self.emit("modeChanged", (new_mode,))

    def set_mode(self, mode):
        self.turret.mode = mode

    def get_mode(self):
        return self.turret.mode

    def get_value(self):
        return self.turret.position

    def set_value(self, position):
        self.turret.position = position

    def get_turret_mnemonic(self):
        return self.turret_name

    def re_emit_values(self):
        self.emit("positionChanged", (self.get_value(),))
        self.emit("modeChanged", (self.get_mode(),))



"""ESRF SC3 Sample Changer Hardware Object"""

import ESRF.ESRFSC3 as ESRFSC3
import SC3

from mxcubecore.TaskUtils import task


class Command:
    def __init__(self, cmd):
        self.cmd = cmd

    def isSpecConnected(self):
        return True

    @task
    def __call__(self, *args, **kwargs):
        self.cmd(*args, **kwargs)


class ESRFMD2SC3(ESRFSC3.ESRFSC3):
    def __init__(self, *args, **kwargs):
        ESRFSC3.ESRFSC3.__init__(self, *args, **kwargs)

    def init(self):
        self.controller = self.get_object_by_role("controller")
        SC3.SC3.init(self)
        self.prepareCentringAfterLoad = True
        # self.prepareCentring = Command(controller.moveToPhase("Centring"))
        # self._moveToLoadingPosition = Command(controller.moveToPhase("Transfer"))
        # self._moveToUnloadingPosition = Command(controller.moveToPhase("Transfer"))

    @task
    def unlockMinidiffMotors(self):
        pass

    @task
    def prepare_centring(self, *args, **kwargs):
        # self.controller.moveToPhase("Centring", wait=True, timeout=1000)
        pass

    @task
    def _moveToLoadingPosition(self, *args, **kwargs):
        # self.controller.moveToPhase("Transfer", wait=True, timeout=10000)
        pass

    @task
    def _moveToUnloadingPosition(self, *args, **kwargs):
        # self.controller.moveToPhase("Transfer", wait=True, timeout=10000)
        pass

    def _getLoadingState(self):
        self.controller._wait_ready(10000)



import logging
import sys


class CalculateFlux:
    def __init__(self, fname=None):
        self.FLUX = {}

    def init(
        self,
        fname="/users/blissadm/local/beamline_control/configuration/calibrated_diodes.dat",
    ):
        self.calib_array = self._load_flux_calibration(fname)

    def _load_flux_calibration(self, fname):
        try:
            f = open(fname)
            array = []
            nb_line = 0
            for line in f:
                if not line.startswith("#") and line.strip():
                    array.append(map(float, line.split()))
                    nb_line += 1
                else:
                    if line.startswith("#"):
                        # line = line[1:]
                        ll = line[1:].split()
                        self.labels = []
                        idx = []
                        for i in ll:
                            self.labels.append(i.lower())
                            idx.append(0)
            f.close()
        except IOError:
            logging.exception("Cannot read calibrated diodes file")
        self.FLUX = dict(zip(self.labels, idx))
        return array

    def calc_flux_coef(self, en):
        if en < 4:
            logging.exception("Cannot calculate flux - unknown energy")
            raise ValueError
        # calculations are made in eV, input might be in KeV
        if en < 1000:
            en *= 1000

        self.FLUX[self.labels[0]] = int(en)

        if en >= self.calib_array[0][0]:
            calib = self.calib_array[0][1:]
        elif en <= self.calib_array[-1][0]:
            calib = self.calib_array[-1][1:]
        else:
            calib = self._interpol(self.calib_array, int(en))

        for i in calib:
            self.FLUX[self.labels[calib.index(i) + 1]] = i

        return calib

    def _interpol(self, arr, val, debug=0):
        larr = []
        for i, vals in enumerate(arr):
            if abs(vals[0] - val) < 10:
                return vals[1:]

            larr.append(abs(vals[0] - val))
        min_index = larr.index(min(larr))
        x1 = arr[min_index][0]
        y1 = arr[min_index][1]
        try:
            z1 = arr[min_index][2]
        except Exception:
            pass
        if debug:
            print(arr[min_index])
        if x1 < val and min_index > 0:
            min_index -= 1
        elif x1 > val and min_index < len(arr) - 1:
            min_index += 1
        else:
            return None
        if debug:
            print(arr[min_index])
        x2 = arr[min_index][0]
        y2 = arr[min_index][1]
        try:
            z2 = arr[min_index][2]
        except Exception:
            pass
        bb = (y2 - y1) / (x2 - x1)
        aa = y1 - bb * x1
        try:
            dd = (z2 - z1) / (x2 - x1)
            cc = z1 - dd * x1
            return [aa + bb * val, cc + dd * val]
        except Exception:
            return aa + bb * val


if __name__ == "__main__":
    fl = CalculateFlux()
    fname = sys.argv[1]
    fl.init(fname)
    en = float(sys.argv[2]) * 1000
    ab = fl.calc_flux_coef(en)
    print(ab)



from mxcubecore.TaskUtils import task

from .ESRFEnergyScan import (
    ESRFEnergyScan,
    TunableEnergy,
)


class ID231EnergyScan(ESRFEnergyScan):
    def __init__(self, name):
        ESRFEnergyScan.__init__(self, name, TunableEnergy())

    @task
    def energy_scan_hook(self, energy_scan_parameters):
        self.energy = energy_scan_parameters["edgeEnergy"]
        if self.energy_scan_parameters["findattEnergy"]:
            ESRFEnergyScan.move_energy(self, energy_scan_parameters["findattEnergy"])

    def calculate_und_gaps(self, energy, undulator="u21d"):
        GAPS = {}
        return GAPS

    @task
    def move_undulators(self, gaps):
        pass

    @task
    def set_mca_roi(self, eroi_min, eroi_max):
        self.execute_command("calculateMcaRoi", eroi_min, eroi_max)

    @task
    def choose_attenuation(self):
        self.execute_command("chooseAttenuation")
        self.energy_scan_parameters["transmissionFactor"] = (
            self.transmission.get_value()
        )

    @task
    def execute_energy_scan(self, energy_scan_parameters):
        self.execute_command("executeScan", energy_scan_parameters)



# encoding: utf-8
#
# License:
#
# This file is part of MXCuBE.
#
# MXCuBE is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# MXCuBE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MXCuBE. If not, see <https://www.gnu.org/licenses/>.

"""General data and functions, that can be shared between different HardwareObjects"""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import re

from scipy.constants import (
    c,
    e,
    h,
)

__date__ = "19/06/17"
__credits__ = ["MXCuBE collaboration"]

# Constants

try:
    # Python 2
    string_types = (basestring,)
    text_type = unicode
    binary_type = str
except Exception:
    # Python 3+
    string_types = (str,)
    text_type = str
    binary_type = bytes

# Conversion from kEv to A, wavelength = HC_OVER_E/energy
HC_OVER_E = h * c / e * 10e6


# Utility functions:


def java_property(keyword, value, quote_value=False):
    """Return argument list for command line invocation setting java property

    keyword, value are stringtypes"""
    if value is None:
        return ["-D" + keyword]
    else:
        if value and quote_value:
            value = quoted_string(value)
        return ["-D%s=%s" % (keyword, value)]


def command_option(keyword, value, prefix="-", quote_value=False):
    """Return argument list for command line option"""
    if value is None:
        return [prefix + keyword]
    else:
        if value and quote_value:
            value = quoted_string(value)
        else:
            value = str(value)
        return [prefix + keyword, value]


def quoted_string(text):
    """Return quoted value of a (single-line) string

    Intended for command line arguments.
    Will work for Python 2 str or unicode, OR Python 3 str and (some) bytes).
    Somewhat fragile, will definitely break for multiline strings
    or strings containing both single and double quotes
    """
    result = ensure_text(text)
    if not '"' in result:
        result = "".join(('"', result, '"'))
    elif not "'" in result:
        result = "".join(("'", result, "'"))
    else:
        result = repr(result)
    ind = 0
    for ind, char in enumerate(result):
        if char in ('"', "'"):
            break
    #
    return result[ind:]


def convert_string_value(text):
    """Convert input string to int, float, or string (in order of priority)"""
    try:
        return int(text)
    except ValueError:
        try:
            return float(text)
        except ValueError:
            return text


# 'Borrowed' from six, pending installation as a dependency
def ensure_text(chars, encoding="utf-8", errors="strict"):
    """Coerce *chars* to six.text_type.
    For Python 2:
      - `unicode` -> `unicode`
      - `str` -> `unicode`
    For Python 3:
      - `str` -> `str`
      - `bytes` -> decoded to `str`
    """
    if isinstance(chars, binary_type):
        return chars.decode(encoding, errors)
    elif isinstance(chars, text_type):
        return chars
    else:
        raise TypeError("not expecting type '%s'" % type(chars))


def make_table(column_names, rows):
    """Generate string with pretty-printed table

    Args:
        column_names (Sequence[str]) : Column names
        rows (Sequence[Sequence[str]]) : List of row data

    Returns:

    """
    lines = []
    longest_cols = [
        (max([len(str(row[i])) for row in rows]) + 3) for i in range(len(rows[0]))
    ]
    longest_cols = list(
        max(longest_cols[ind], len(txt)) for ind, txt in enumerate(column_names)
    )
    ruler = "+" + "=" * sum(longest_cols) + "+"
    row_format = "| ".join(
        ["{:<" + str(longest_col) + "}" for longest_col in longest_cols]
    )

    lines.append(ruler)
    lines.append("| %s" % row_format.format(*column_names))
    lines.append(ruler)

    for row in rows:
        lines.append("| %s" % row_format.format(*row))
    lines.append(ruler)
    #
    return "\n".join(lines)


def camel_to_snake(name):
    pattern = re.compile(r"(?<!^)(?=[A-Z])")
    return pattern.sub("_", name).lower()



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.dispatcher import dispatcher
from mxcubecore.queue_entry.base_queue_entry import (
    QUEUE_ENTRY_STATUS,
    BaseQueueEntry,
    QueueAbortedException,
    QueueExecutionException,
)

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class EnergyScanQueueEntry(BaseQueueEntry):
    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)
        self.energy_scan_task = None
        self._failed = False

    def __getstate__(self):
        d = dict(self.__dict__)
        d["energy_scan_task"] = None
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)

    def execute(self):
        BaseQueueEntry.execute(self)

        if HWR.beamline.energy_scan:
            energy_scan = self.get_data_model()
            self.get_view().setText(1, "Starting energy scan")

            if energy_scan.shape is not None:
                point = HWR.beamline.sample_view.get_shape(energy_scan.shape)
                energy_scan.centred_position = point.get_centred_position()

            sample_model = self.get_data_model().get_sample_node()

            sample_lims_id = sample_model.lims_id

            # No sample id, pass None to start_energy_scan
            if sample_lims_id == -1:
                sample_lims_id = None

            self.energy_scan_task = gevent.spawn(
                HWR.beamline.energy_scan.start_energy_scan,
                energy_scan.element_symbol,
                energy_scan.edge,
                energy_scan.path_template.directory,
                energy_scan.path_template.get_prefix(),
                HWR.beamline.session.session_id,
                sample_lims_id,
                cpos=energy_scan.centred_position,
            )

        HWR.beamline.energy_scan.ready_event.wait()
        HWR.beamline.energy_scan.ready_event.clear()

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)
        self._failed = False

        qc = self.get_queue_controller()

        qc.connect(
            HWR.beamline.energy_scan,
            "scanStatusChanged",
            self.energy_scan_status_changed,
        )

        qc.connect(
            HWR.beamline.energy_scan, "energyScanStarted", self.energy_scan_started
        )

        qc.connect(
            HWR.beamline.energy_scan, "energyScanFinished", self.energy_scan_finished
        )

        qc.connect(
            HWR.beamline.energy_scan, "energyScanFailed", self.energy_scan_failed
        )

    def post_execute(self):
        BaseQueueEntry.post_execute(self)
        qc = self.get_queue_controller()

        qc.disconnect(
            HWR.beamline.energy_scan,
            "scanStatusChanged",
            self.energy_scan_status_changed,
        )

        qc.disconnect(
            HWR.beamline.energy_scan, "energyScanStarted", self.energy_scan_started
        )

        qc.disconnect(
            HWR.beamline.energy_scan, "energyScanFinished", self.energy_scan_finished
        )

        qc.disconnect(
            HWR.beamline.energy_scan, "energyScanFailed", self.energy_scan_failed
        )

        if self._failed:
            raise QueueAbortedException("Queue stopped", self)
        self.get_view().set_checkable(False)

    def energy_scan_status_changed(self, msg):
        logging.getLogger("user_level_log").info(msg)

    def energy_scan_started(self, *args):
        logging.getLogger("user_level_log").info("Energy scan started.")
        self.get_view().setText(1, "In progress")

    def energy_scan_finished(self, scan_info):
        self.get_view().setText(1, "Done")

        energy_scan = self.get_data_model()

        (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        ) = HWR.beamline.energy_scan.do_chooch(
            energy_scan.element_symbol,
            energy_scan.edge,
            energy_scan.path_template.directory,
            energy_scan.path_template.get_archive_directory(),
            "%s_%d"
            % (
                energy_scan.path_template.get_prefix(),
                energy_scan.path_template.run_number,
            ),
        )
        # scan_file_archive_path,
        # scan_file_path)

        # Trying to get the sample from the EnergyScan model instead through
        # the view. Keeping the old way fore backward compatability
        if energy_scan.sample:
            sample = energy_scan.sample
        else:
            sample = self.get_view().parent().parent().get_model()

        sample.crystals[0].energy_scan_result.peak = pk
        sample.crystals[0].energy_scan_result.inflection = ip
        sample.crystals[0].energy_scan_result.first_remote = rm
        sample.crystals[0].energy_scan_result.second_remote = None

        energy_scan.result.pk = pk
        energy_scan.result.fppPeak = fppPeak
        energy_scan.result.fpPeak = fpPeak
        energy_scan.result.ip = ip
        energy_scan.result.fppInfl = fppInfl
        energy_scan.result.fpInfl = fpInfl
        energy_scan.result.rm = rm
        energy_scan.result.chooch_graph_x = chooch_graph_x
        energy_scan.result.chooch_graph_y1 = chooch_graph_y1
        energy_scan.result.chooch_graph_y2 = chooch_graph_y2
        energy_scan.result.title = title
        try:
            energy_scan.result.data = HWR.beamline.energy_scan.get_scan_data()
        except Exception:
            pass

        if (
            sample.crystals[0].energy_scan_result.peak
            and sample.crystals[0].energy_scan_result.inflection
        ):
            logging.getLogger("user_level_log").info(
                "Energy scan: Result peak: %.4f, inflection: %.4f"
                % (
                    sample.crystals[0].energy_scan_result.peak,
                    sample.crystals[0].energy_scan_result.inflection,
                )
            )

        self.get_view().setText(1, "Done")
        self._queue_controller.emit("energy_scan_finished", (pk, ip, rm, sample))

    def energy_scan_failed(self):
        self._failed = True
        self.get_view().setText(1, "Failed")
        self.status = QUEUE_ENTRY_STATUS.FAILED
        logging.getLogger("user_level_log").error("Energy scan: failed")
        raise QueueExecutionException("Energy scan failed", self)

    def stop(self):
        BaseQueueEntry.stop(self)

        try:
            # self.get_view().setText(1, 'Stopping ...')
            HWR.beamline.energy_scan.cancelEnergyScan()

            if self.centring_task:
                self.centring_task.kill(block=False)
        except gevent.GreenletExit:
            raise

        self.get_view().setText(1, "Stopped")
        logging.getLogger("queue_exec").info("Calling stop on: " + str(self))
        # this is to work around the remote access problem
        dispatcher.send("collect_finished")
        raise QueueAbortedException("Queue stopped", self)

    def get_type_str(self):
        return "Energy scan"



import logging

import gevent
import gevent.monkey
import numpy
from dispatcher import saferef
from gevent import _threading
from gevent.event import Event

try:
    import Queue as queue
except ImportError:
    import queue


log = logging.getLogger("HWR")

POLLERS = {}

gevent_version = list(map(int, gevent.__version__.split(".")))


class _NotInitializedValue:
    pass


NotInitializedValue = _NotInitializedValue()


class PollingException:
    def __init__(self, e, poller_id):
        self.original_exception = e
        self.poller_id = poller_id


def get_poller(poller_id):
    return POLLERS.get(poller_id)


def poll(
    polled_call,
    polled_call_args=(),
    polling_period=1000,
    value_changed_callback=None,
    error_callback=None,
    compare=True,
    start_delay=0,
    start_value=NotInitializedValue,
):
    for poller in POLLERS.values():
        poller_polled_call = poller.polled_call_ref()
        if poller_polled_call == polled_call and poller.args == polled_call_args:
            poller.set_polling_period(min(polling_period, poller.get_polling_period()))
            return poller

    poller = _Poller(
        polled_call,
        polled_call_args,
        polling_period,
        value_changed_callback,
        error_callback,
        compare,
    )
    poller.old_res = start_value
    POLLERS[poller.get_id()] = poller
    poller.start_delayed(start_delay)
    return poller


class _Poller:
    def __init__(
        self,
        polled_call,
        polled_call_args=(),
        polling_period=1000,
        value_changed_callback=None,
        error_callback=None,
        compare=True,
    ):
        self.polled_call = polled_call
        self.polled_call_ref = saferef.safe_ref(polled_call)
        self.args = polled_call_args
        self.polling_period = polling_period
        self.value_changed_callback_ref = saferef.safe_ref(value_changed_callback)
        self.error_callback_ref = saferef.safe_ref(error_callback)
        self.compare = compare
        self.old_res = NotInitializedValue
        self.queue = queue.Queue()
        self.delay = 0
        self.stop_event = Event()
        self.async_watcher = gevent.get_hub().loop.async_()

    def start_delayed(self, delay):
        self.delay = delay
        _threading.start_new_thread(self.run, ())

    def stop(self):
        self.stop_event.set()
        del POLLERS[self.get_id()]

    def is_stopped(self):
        return self.stop_event.is_set()

    def get_id(self):
        return id(self)

    def get_polling_period(self):
        return self.polling_period

    def set_polling_period(self, polling_period):
        self.polling_period = polling_period

    def restart(self, delay=0):
        self.stop()

        polled_call = self.polled_call_ref()
        value_changed_cb = self.value_changed_callback_ref()
        error_cb = self.error_callback_ref()
        if polled_call is not None:
            return poll(
                polled_call,
                self.args,
                self.polling_period,
                value_changed_cb,
                error_cb,
                self.compare,
                delay,
                start_value=self.old_res,
            )

    def new_event(self):
        while True:
            try:
                res = self.queue.get_nowait()
            except queue.Empty:
                break

            if isinstance(res, PollingException):
                cb = self.error_callback_ref()
                if cb is not None:
                    gevent.spawn(cb, res.original_exception, res.poller_id)
            else:
                cb = self.value_changed_callback_ref()
                if cb is not None:
                    gevent.spawn(cb, res)

    def run(self):
        sleep = gevent.monkey.get_original("time", "sleep")

        self.async_watcher.start(self.new_event)

        err_callback_args = None
        error_cb = None
        first_run = True

        while not self.stop_event.is_set():
            if first_run and self.delay:
                sleep(self.delay / 1000.0)
            first_run = False

            if self.stop_event.is_set():
                break

            polled_call = self.polled_call_ref()
            if polled_call is None:
                break

            try:
                res = polled_call(*self.args)
            except Exception as e:
                if self.stop_event.is_set():
                    break
                error_cb = self.error_callback_ref()
                if error_cb is not None:
                    self.queue.put(PollingException(e, self.get_id()))
                self.old_res = NotInitializedValue
                break

            del polled_call

            if self.stop_event.is_set():
                break

            if isinstance(res, numpy.ndarray):  # for arrays
                comparison = res == self.old_res
                if isinstance(comparison, bool):
                    is_equal = comparison
                else:
                    is_equal = all(comparison)
            else:
                is_equal = res == self.old_res

            if self.compare and is_equal:
                # do nothing: previous value is the same as "new" value
                pass
            else:
                new_value = True
                if self.compare:
                    new_value = not is_equal

                if new_value:
                    self.old_res = res
                    self.queue.put(res)
                    self.async_watcher.send()

            sleep(self.polling_period / 1000.0)

        if error_cb is not None:
            self.async_watcher.send()



import logging
import math
import os
import os.path
import shutil

# import PyChooch
# to run chooch in shell
import subprocess
import time

import gevent
import numpy
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractEnergyScan import AbstractEnergyScan
from mxcubecore.TaskUtils import task


class GetStaticParameters:
    def __init__(self, config_file, element, edge):
        self.element = element
        self.edge = edge
        self.pars_dict = {}
        self.pars_dict = self._read_from_file(config_file)

    def _read_from_file(self, config_file):
        with open(config_file, "r") as f:
            array = []
            for line in f:
                if not line.startswith("#") and self.element in line:
                    array = line.split()
                    break

            try:
                static_pars = {}
                static_pars["atomic_nb"] = int(array[0])
                static_pars["eroi_min"] = float(array[11]) / 1000.0
                static_pars["eroi_max"] = float(array[12]) / 1000.0

                if "K" in self.edge:
                    th_energy = float(array[3]) / 1000.0
                else:
                    if "1" in self.edge:
                        # L1
                        th_energy = float(array[6]) / 1000.0
                    elif "2" in self.edge:
                        # L2
                        th_energy = float(array[7]) / 1000.0
                    else:
                        # L or L3
                        th_energy = float(array[8]) / 1000.0

                # all the values are in keV
                static_pars["edgeEnergy"] = th_energy
                static_pars["startEnergy"] = th_energy - 0.05
                static_pars["endEnergy"] = th_energy + 0.05
                static_pars["findattEnergy"] = th_energy + 0.03
                static_pars["remoteEnergy"] = th_energy + 1
                return static_pars
            except Exception as err:
                print(err)
                return {}


class ESRFEnergyScan(AbstractEnergyScan):
    def __init__(self, name):
        super().__init__(name)
        # self._tunable_bl = tunable_bl
        self.ctrl = None

    def execute_command(self, command_name, *args, **kwargs):
        wait = kwargs.get("wait", True)
        cmd_obj = self.get_command_object(command_name)
        return cmd_obj(*args, wait=wait)

    def init(self):
        self.energy_obj = HWR.beamline.energy
        self.beamsize = self.get_object_by_role("beamsize")
        self.ctrl = self.get_object_by_role("controller")
        self.ready_event = gevent.event.Event()
        if HWR.beamline.lims is None:
            logging.getLogger("HWR").warning(
                "EnergyScan: you should specify the database hardware object"
            )
        self.scanInfo = None
        #self._tunable_bl.energy_obj = self.energy_obj

    def is_connected(self):
        return True

    def get_static_parameters(self, config_file, element, edge):
        pars = GetStaticParameters(config_file, element, edge).pars_dict

        offset_keV = self.get_property("offset_keV")
        pars["startEnergy"] += offset_keV
        pars["endEnergy"] += offset_keV
        pars["element"] = element

        return pars

    def escan_prepare(self):

        if self.beamsize:
            bsX = self.beamsize.get_size(self.beamsize.get_value().name)
            self.energy_scan_parameters["beamSizeHorizontal"] = bsX
            self.energy_scan_parameters["beamSizeVertical"] = bsX

    def escan_postscan(self):
        self.ctrl.diffractometer.fldet_out()
        #self.ctrl.fluodet.OUT

    def escan_cleanup(self):
        self.close_fast_shutter()
        HWR.beamline.safety_shutter.close()
        self.emit("energyScanFailed", ())
        self.ready_event.set()

    def close_fast_shutter(self):
        self.ctrl.diffractometer.msclose()

    def open_fast_shutter(self):
        self.ctrl.diffractometer.msopen()

    def cancelEnergyScan(self, *args):
        """ Called by queue_entry.py. To be removed"""
        self.escan_cleanup()

    def storeEnergyScan(self):
        if HWR.beamline.lims is None:
            return
        try:
            int(self.energy_scan_parameters["sessionId"])
        except Exception:
            return

        # remove unnecessary for ISPyB fields:
        self.energy_scan_parameters.pop("prefix")
        self.energy_scan_parameters.pop("eroi_min")
        self.energy_scan_parameters.pop("eroi_max")
        self.energy_scan_parameters.pop("findattEnergy")
        self.energy_scan_parameters.pop("edge")
        self.energy_scan_parameters.pop("directory")
        self.energy_scan_parameters.pop("atomic_nb")

        gevent.spawn(
            StoreEnergyScanThread, HWR.beamline.lims, self.energy_scan_parameters
        )

    def do_chooch(self, elt, edge, directory, archive_directory, prefix):
        self.energy_scan_parameters["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")

        raw_data_file = os.path.join(directory, "data.raw")

        symbol = "_".join((elt, edge))
        archive_prefix = "_".join((prefix, symbol))
        raw_scan_file = os.path.join(directory, (archive_prefix + ".raw"))
        efs_scan_file = raw_scan_file.replace(".raw", ".efs")
        raw_arch_file = os.path.join(archive_directory, (archive_prefix + "1" + ".raw"))

        i = 0
        while os.path.isfile(raw_arch_file):
            i += 1
            raw_arch_file = os.path.join(
                archive_directory, (archive_prefix + str(i) + ".raw")
            )

        png_scan_file = raw_scan_file.replace(".raw", ".png")
        png_arch_file = raw_arch_file.replace(".raw", ".png")

        if not os.path.exists(archive_directory):
            os.makedirs(archive_directory)
        try:
            f = open(raw_scan_file, "w")
        except IOError:
            self.storeEnergyScan()
            self.emit("energyScanFailed", ())
            return
        else:
            scan_data = []
            try:
                with open(raw_data_file, "r") as raw_file:
                    for line in raw_file.readlines()[2:]:
                        try:
                            (x, y) = line.split("\t")
                        except Exception:
                            (x, y) = line.split()
                        x = float(x.strip())
                        y = float(y.strip())
                        scan_data.append((x, y))
                        f.write("%f,%f\r\n" % (x, y))
                f.close()
            except IOError:
                self.storeEnergyScan()
                self.emit("energyScanFailed", ())
                return

        shutil.copy2(raw_scan_file, raw_arch_file)
        self.energy_scan_parameters["scanFileFullPath"] = raw_arch_file

        """
        result = PyChooch.calc(scan_data, elt, edge, efs_scan_file)
        # PyChooch occasionally returns an error and the result
        # the sleep command assures that we get the result
        time.sleep(1)
        print(result[0])
        pk = result[0] / 1000.0
        fppPeak = result[1]
        fpPeak = result[2]
        ip = result[3] / 1000.0
        fppInfl = result[4]
        fpInfl = result[5]
        chooch_graph_data = result[6]
        """
        # while waiting fro chooch to work...
        subprocess.call(
            [
                "/cvmfs/sb.esrf.fr/bin/chooch",
                "-e",
                elt,
                "-a",
                edge,
                "-o",
                efs_scan_file,
                "-g",
                png_scan_file,
                raw_data_file,
            ]
        )
        time.sleep(5)
        with open(efs_scan_file, "r") as f:
            for _ in range(3):
                next(f)
            nparr = numpy.array([list(map(float, line.split())) for line in f])
        fppPeak = nparr[:, 1].max()
        idx = numpy.where(nparr[:, 1] == fppPeak)
        pk = nparr[:, 0][idx][0] / 1000.0
        fpPeak = nparr[:, 2][idx][0]
        fppInfl = nparr[:, 2].min()
        idx = numpy.where(nparr[:, 2] == fppInfl)
        ip = nparr[:, 0][idx][0] / 1000.0
        fpInfl = nparr[:, 1][idx][0]
        rm = pk + 0.03

        comm = ""
        th_edge = float(self.energy_scan_parameters["edgeEnergy"])

        logging.getLogger("HWR").info(
            "Chooch results: pk = %f, ip = %f, rm = %f, Theoretical edge: %f"
            % (pk, ip, rm, th_edge)
        )

        # +- shift from the theoretical edge [eV]
        edge_shift = 10
        calc_shift = (th_edge - ip) * 1000
        if math.fabs(calc_shift) > edge_shift:
            rm = th_edge + 0.03
            comm = "%s" % "below" if (calc_shift) > edge_shift else "above"
            comm = (
                "Calculated peak (%f) is more than %d eV %s the theoretical value (%f)."
                % (pk, edge_shift, comm, th_edge)
            )

            logging.getLogger("user_level_log").info(
                "EnergyScan: %s Check your scan and choose the energies manually" % comm
            )
            pk = 0
            ip = 0

        efs_arch_file = raw_arch_file.replace(".raw", ".efs")
        if os.path.isfile(efs_scan_file):
            shutil.copy2(efs_scan_file, efs_arch_file)
        else:
            self.storeEnergyScan()
            self.emit("energyScanFailed", ())
            return

        self.energy_scan_parameters["filename"] = raw_arch_file.split("/")[-1]
        self.energy_scan_parameters["peakEnergy"] = pk
        self.energy_scan_parameters["inflectionEnergy"] = ip
        self.energy_scan_parameters["remoteEnergy"] = rm
        self.energy_scan_parameters["peakFPrime"] = fpPeak
        self.energy_scan_parameters["peakFDoublePrime"] = fppPeak
        self.energy_scan_parameters["inflectionFPrime"] = fpInfl
        self.energy_scan_parameters["inflectionFDoublePrime"] = fppInfl
        self.energy_scan_parameters["comments"] = comm

        """
        chooch_graph_x, chooch_graph_y1, chooch_graph_y2 = zip(*chooch_graph_data)
        chooch_graph_x = list(chooch_graph_x)
        chooch_graph_x = [x / 1000.0 for x in chooch_graph_x]
        """
        logging.getLogger("HWR").info("Saving png")
        # prepare to save png files
        title = "%10s  %6s  %6s\n%10s  %6.2f  %6.2f\n%10s  %6.2f  %6.2f" % (
            "energy",
            "f'",
            "f''",
            pk,
            fpPeak,
            fppPeak,
            ip,
            fpInfl,
            fppInfl,
        )
        if os.path.isfile(png_scan_file):
            shutil.copy2(png_scan_file, png_arch_file)
        else:
            self.storeEnergyScan()
            self.emit("energyScanFailed", ())
            return
        """
        fig = Figure(figsize=(15, 11))
        ax = fig.add_subplot(211)
        ax.set_title("%s\n%s" % (efs_scan_file, title))
        ax.grid(True)
        ax.plot(*(zip(*scan_data)), **{"color": "black"})
        ax.set_xlabel("Energy")
        ax.set_ylabel("MCA counts")
        ax2 = fig.add_subplot(212)
        ax2.grid(True)
        ax2.set_xlabel("Energy")
        ax2.set_ylabel("")
        handles = []
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y1, color="blue"))
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y2, color="red"))
        canvas = FigureCanvasAgg(fig)
        """
        self.energy_scan_parameters["jpegChoochFileFullPath"] = str(png_arch_file)
        """
        try:
            logging.getLogger("HWR").info(
                "Rendering energy scan and Chooch graphs to PNG file : %s",
                png_scan_file,
            )
            canvas.print_figure(png_scan_file, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not print figure")
        try:
            logging.getLogger("HWR").info(
                "Saving energy scan to archive directory for ISPyB : %s", png_arch_file
            )
            canvas.print_figure(png_arch_file, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not save figure")
        """
        self.storeEnergyScan()

        self.emit(
            "chooch_finished",
            (
                pk,
                fppPeak,
                fpPeak,
                ip,
                fppInfl,
                fpInfl,
                rm,
                # chooch_graph_x,
                # chooch_graph_y1,
                # chooch_graph_y2,
                title,
            ),
        )
        return (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            [],
            [],
            [],
            # chooch_graph_x,
            # chooch_graph_y1,
            # chooch_graph_y2,
            title,
        )


def StoreEnergyScanThread(db_conn, scan_info):
    scan_info = dict(scan_info)
    blsample_id = scan_info["blSampleId"]
    scan_info.pop("blSampleId")

    try:
        db_status = db_conn.store_energy_scan(scan_info)
        if blsample_id is not None:
            try:
                escan_id = int(db_status["energyScanId"])
            except (NameError, KeyError):
                pass
            else:
                asso = {"blSampleId": blsample_id, "energyScanId": escan_id}
                db_conn.associate_bl_sample_and_energy_scan(asso)
    except Exception as e:
        print(e)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""BlissShutter class - interface for shutter controlled by BLISS
Implements _set_value, get_value methods
Bliss states are: UNKNOWN, OPEN, CLOSED, FAULT
"MOVING", "DISABLE", "STANDBY", "RUNNING"
Example yml configuration:

.. code-block:: yaml

 class: BlissShutter.BlissShutter
 configuration:
   actuator_name: safshut
   type: tango
   username: Safety shutter
 objects:
   controller: bliss.yaml
"""

from enum import (
    Enum,
    unique,
)

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@unique
class BlissShutterStates(Enum):
    """Shutter states definitions."""

    OPEN = HardwareObjectState.READY, "OPEN"
    CLOSED = HardwareObjectState.READY, "CLOSED"
    MOVING = HardwareObjectState.BUSY, "MOVING"
    DISABLE = HardwareObjectState.WARNING, "DISABLE"
    AUTOMATIC = HardwareObjectState.READY, "RUNNING"
    UNKNOWN = HardwareObjectState.UNKNOWN, "RUNNING"
    FAULT = HardwareObjectState.WARNING, "FAULT"


class BlissShutter(AbstractShutter):
    """BLISS implementation of AbstractShutter"""

    SPECIFIC_STATES = BlissShutterStates

    def __init__(self, name):
        super().__init__(name)
        self._bliss_obj = None
        self.shutter_type = None
        self.opening_mode = None

    def init(self):
        """Initilise the predefined values"""
        self.controller = self.get_object_by_role("controller")
        super().init()
        self._bliss_obj = getattr(self.controller, self.actuator_name)
        # for now we only treat tango type shutter
        self.shutter_type = self.get_property("type", "tango")
        try:
            if self._bliss_obj.frontend:
                self.opening_mode = self._bliss_obj.mode
        except AttributeError:
            # there is no frontend property
            pass
        if self.shutter_type == "tango":
            self._initialise_values()
        self._poll_task = gevent.spawn(self._poll_state)

        self.update_state()

    def _poll_state(self):
        while True:
            self.update_value(self.get_value())
            gevent.sleep(0.5)

    def _initialise_values(self):
        """Add the tango states to VALUES"""
        values_dict = {item.name: item.value for item in self.VALUES}
        values_dict.update(
            {
                "MOVING": "MOVING",
                "DISABLE": "DISABLE",
                "STANDBY": "STANDBY",
                "FAULT": "FAULT",
            }
        )
        self.VALUES = Enum("ValueEnum", values_dict)

    def get_state(self):
        """Get the device state.
        Returns:
            (enum 'HardwareObjectState'): Device state.
        """
        try:
            _state = self._bliss_obj.state.name
        except (AttributeError, KeyError):
            return self.STATES.UNKNOWN
        return self.SPECIFIC_STATES[_state].value[0]

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): Enum member, corresponding to the value or UNKNOWN.
        """
        # the return from BLISS value is an Enum
        _val = self._bliss_obj.state.name
        return self.value_to_enum(_val)

    def _set_value(self, value):
        if value.name == "OPEN":
            self._bliss_obj.open()
        elif value.name == "CLOSED":
            self._bliss_obj.close()

    def set_mode(self, value):
        """Set automatic or manual mode for a Frontend shutter
        Args:
            value (str): MANUAL or AUTOMATIC
        Raises: NotImplementedError: Not a Fronend shutter.
        """
        self._bliss_obj.mode = value
        self.opening_mode = value



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
import json
import logging
from enum import (
    Enum,
    unique,
)

import gevent
import redis

from mxcubecore.BaseHardwareObjects import HardwareObject


@unique
class PlotType(Enum):
    """
    Defines default plot
    """

    SCATTER = "scatter"


class PlotDim(Enum):
    """
    Defines data dimension
    """

    ONE_D = 1
    TWO_D = 2


class DataType(Enum):
    """
    Defines avialable data types
    """

    FLOAT = "float"


class FrameType(Enum):
    """
    Enum defining the message frame types
    """

    DATA = "data"
    START = "start"
    STOP = "stop"


def one_d_data(x, y):
    """
    Convenience function for creating x, y data
    """
    return {"x": x, "y": y}


def two_d_data(x, y):
    """
    Convenience function for creating x, y, z data
    """
    return {"x": x, "y": y, "z": z}


class DataPublisher(HardwareObject):
    """
    DataPublisher handles data publishing
    """

    def __init__(self, name):
        super(DataPublisher, self).__init__(name)
        self._r = None
        self._subsribe_task = None

    def init(self):
        """
        FWK2 Init method
        """
        super(DataPublisher, self).init()

        rhost = self.get_property("host", "localhost")
        rport = self.get_property("port", 6379)
        rdb = self.get_property("db", 11)

        self._r = redis.Redis(
            host=rhost, port=rport, db=rdb, encoding="utf-8", decode_responses=True
        )

        if not self._subsribe_task:
            self._subsribe_task = gevent.spawn(self._handle_messages)

    def _handle_messages(self):
        """
        Listens for published data and handles the data.
        """
        pubsub = self._r.pubsub(ignore_subscribe_messages=True)
        pubsub.psubscribe("HWR_DP_NEW_DATA_POINT_*")

        _data = {}

        # The descriptions of active sources for fast access
        # while publishing data
        active_source_desc = {}

        for message in pubsub.listen():
            if message:
                try:
                    redis_channel = message["channel"]
                    _id = redis_channel.split("_")[-1]

                    data = json.loads(message["data"])

                    if data["type"] == FrameType.START.value:
                        _data[redis_channel] = {"x": [], "y": []}

                        self._update_description(_id, {"running": True})

                        # Clear previous data so that we are not acumelating
                        # with previously published data
                        self._clear_data(_id)

                        self.emit(
                            "start", self.get_description(_id, include_data=True)[0]
                        )

                        active_source_desc[redis_channel] = self._get_description(_id)

                    elif data["type"] == FrameType.STOP.value:
                        self._update_description(_id, {"running": False})
                        self.emit(
                            "end", self.get_description(_id, include_data=True)[0]
                        )
                        active_source_desc.pop(redis_channel)
                    elif data["type"] == FrameType.DATA.value:
                        _data[redis_channel] = {
                            "x": _data[redis_channel]["x"] + [data["data"]["x"]],
                            "y": _data[redis_channel]["y"] + [data["data"]["y"]],
                        }

                        self.emit(
                            "data",
                            {"id": _id, "data": data["data"]},
                        )

                        self._append_data(
                            _id, data["data"], active_source_desc[redis_channel]
                        )
                    else:
                        msg = "Unknown frame type %s" % message
                        logging.getLogger("HWR").error(msg)
                except Exception:
                    msg = "Could not parse data in %s" % message
                    logging.getLogger("HWR").exception(msg)

    def _remove_available(self, _id):
        """
        Remove source with _id from list of avialable sources

        Args:
            _id (str): The id of the source to remove
        """
        sources = self._get_available()
        sources = {} if not sources else sources
        sources.pop(_id)

        self._r.set("HWR_DP_PUBLISHERS", json.dumps(sources))

    def _add_avilable(self, _id):
        """
        Add source with _id to list of avialable sources

        Args:
            _id (str): The id of the sources to remove
        """
        sources = self._get_available()
        sources = {} if not sources else sources
        sources[_id] = "HWR_DP_NEW_DATA_POINT_%s" % _id

        self._r.set("HWR_DP_PUBLISHERS", json.dumps(sources))

    def _get_available(self):
        """
        Returns:
            (dict): Where the key is the id of source and the value
                    the channel name to publish data to.
        """
        sources = self._r.get("HWR_DP_PUBLISHERS")
        sources = json.loads(sources) if sources else {}

        return sources

    def _set_description(self, _id, desc):
        """
        Sets the description of source with _id to desc

        Args:
            _id (str): The id of the source to remove
            desc (dict): "id": str,
                         "channel": str,
                         "name": str,
                         "data_type": str
                         "data_dim": float
                         "plot_type": str
                         "sample_rate": float
                         "content_type": str
                         "range": list (min, max)
                         "meta": str
                         "running": boolean,
        """
        self._r.set("HWR_DP_%s_DESCRIPTION" % _id, json.dumps(desc))

    def _get_description(self, _id):
        """
        Return the description of source with _id

        Returns:
            (dict): "id": str,
                     "channel": str,
                     "name": str,
                     "data_type": str
                     "data_dim": float
                     "plot_type": str
                     "sample_rate": float
                     "content_type": str
                     "range": list (min, max)
                     "meta": str
                     "running": boolean,
        """
        return json.loads(self._r.get("HWR_DP_%s_DESCRIPTION" % _id))

    def _update_description(self, _id, data):
        """
        Update the description of source with _id with data
        Args:
            _id (str): The id of the source to remove
            desc (dict): with key, value pairs to update
        """
        desc = self._get_description(_id)
        desc.update(data)
        self._set_description(_id, desc)

    def _append_data(self, _id, data, desc):
        """
        Append data to source with _id

        Args:
            _id (str): The id of the source to remove
            desc (dict): Publisher description
            data: x, y, (z) data to append
        """
        self._r.rpush("HWR_DP_%s_DATA_X" % _id, data.get("x", float("nan")))
        self._r.rpush("HWR_DP_%s_DATA_Y" % _id, data.get("y", float("nan")))

        if desc["data_dim"] > 1:
            self._r.rpush("HWR_DP_%s_DATA_Z" % _id, data.get("z", float("nan")))

    def _clear_data(self, _id):
        """
        Clear data of source with _id
        """
        desc = self._get_description(_id)

        self._r.delete("HWR_DP_%s_DATA_X" % _id)
        self._r.delete("HWR_DP_%s_DATA_Y" % _id)

        if desc["data_dim"] > 1:
            self._r.delete("HWR_DP_%s_DATA_Z" % _id)

    def _publish(self, _id, data):
        """
        Publish data to source with _id

        Args:
            _id (str): The id of the source to remove
            data: x, y, (z) data to append
        """
        self._r.publish("HWR_DP_NEW_DATA_POINT_%s" % _id, json.dumps(data))

    def register(
        self,
        _id,
        name,
        channel,
        axis_labels=["x", "y", "z"],
        data_type=DataType.FLOAT,
        data_dim=PlotDim.ONE_D,
        plot_type=PlotType.SCATTER,
        content_type="",
        sample_rate=0.5,
        _range=(None, None),
        meta={},
    ):
        plot_description = {
            "id": _id,
            "name": name,
            "axis_labels": axis_labels,
            "channel": channel,
            "data_type": data_type.value,
            "data_dim": data_dim.value,
            "plot_type": plot_type.value,
            "sample_rate": sample_rate,
            "content_type": content_type,
            "range": _range,
            "meta": meta,
            "running": False,
        }

        self._set_description(_id, plot_description)
        self._add_avilable(_id)

        return _id

    def pub(self, _id, data):
        self._publish(_id, {"type": FrameType.DATA.value, "data": data})

    def start(self, _id):
        self._publish(_id, {"type": FrameType.START.value, "data": {}})

    def stop(self, _id):
        self._update_description(_id, {"running": False})
        self._publish(_id, {"type": FrameType.STOP.value, "data": {}})

    def get_description(self, _id=None, include_data=False):
        desc = []

        if _id:
            _d = self._get_description(_id)

            if include_data:
                _d.update({"values": self.get_data(_id)})

            desc = [_d]

        else:
            available = self._get_available()

            for _id in available.keys():
                _d = self._get_description(_id)

            if include_data:
                _d.update({"values": self.get_data(_id)})

            desc.append(_d)

        return desc

    def get_data(self, _id):
        desc = self._get_description(_id)
        data = {
            "x": self._r.lrange("HWR_DP_%s_DATA_X" % _id, 0, -1),
            "y": self._r.lrange("HWR_DP_%s_DATA_Y" % _id, 0, -1),
        }

        if desc["data_dim"] > 1:
            data.update(
                {
                    "z": self._r.lrange("HWR_DP_%s_DATA_Z" % _id, 0, -1),
                }
            )

        return data
